# Copyright 2024 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Implement a multiprocess PPOCritic
"""

import itertools

import torch
import torch.distributed
from torch import nn, optim
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP

from siirl import DataProto
from siirl.workers.dag_worker import core_algos
from siirl.utils.extras.device import get_device_id, get_device_name, is_cuda_available, is_npu_available
from siirl.utils.model_utils.fsdp_utils import FSDPModule, fsdp2_clip_grad_norm_
from siirl.utils.extras.py_functional import append_to_dict
from siirl.utils.model_utils.seqlen_balancing import get_reverse_idx, rearrange_micro_batches
from siirl.utils.model_utils.torch_functional import masked_mean
from siirl.utils.model_utils.ulysses import gather_outpus_and_unpad, ulysses_pad_and_slice_inputs
from siirl.workers.critic import BasePPOCritic
from siirl.utils.params.model_args import CriticArguments

if is_cuda_available:
    from flash_attn.bert_padding import index_first_axis, pad_input, rearrange, unpad_input
elif is_npu_available:
    from transformers.integrations.npu_flash_attention import index_first_axis, pad_input, rearrange, unpad_input


class DataParallelPPOCritic(BasePPOCritic):
    def __init__(self, config: CriticArguments, critic_module: nn.Module, critic_optimizer: optim.Optimizer):
        super().__init__(config=config)
        self.critic_module = critic_module
        self.critic_optimizer = critic_optimizer
        self.use_remove_padding = self.config.model.use_remove_padding
        print(f"Critic use_remove_padding={self.use_remove_padding}")

        self.ulysses_sequence_parallel_size = self.config.ulysses_sequence_parallel_size
        self.device_name = get_device_name()

    def _forward_micro_batch(self, micro_batch):
        response_length = micro_batch["responses"].size(-1)
        multi_modal_inputs = {}
        if "multi_modal_inputs" in micro_batch.keys():
            for key in micro_batch["multi_modal_inputs"][0].keys():
                multi_modal_inputs[key] = torch.cat([inputs[key] for inputs in micro_batch["multi_modal_inputs"]], dim=0)

        with torch.autocast(device_type=self.device_name, dtype=torch.bfloat16):
            input_ids = micro_batch["input_ids"]
            batch, seqlen = input_ids.shape
            attention_mask = micro_batch["attention_mask"]
            position_ids = micro_batch["position_ids"]
            if position_ids.dim() == 3:  # qwen2vl mrope
                position_ids = position_ids.transpose(0, 1)

            if self.use_remove_padding:
                input_ids_rmpad, indices, *_ = unpad_input(input_ids.unsqueeze(-1), attention_mask)  # input_ids_rmpad (total_nnz, ...)
                input_ids_rmpad = input_ids_rmpad.transpose(0, 1)  # (1, total_nnz)

                # unpad the position_ids to align the rotary
                if position_ids.dim() == 3:
                    position_ids_rmpad = index_first_axis(rearrange(position_ids, "c b s ... -> (b s) c ..."), indices).transpose(0, 1).unsqueeze(1)  # (3, bsz, seqlen) -> (3, 1, bsz * seqlen)
                else:
                    position_ids_rmpad = index_first_axis(rearrange(position_ids.unsqueeze(-1), "b s ... -> (b s) ..."), indices).transpose(0, 1)

                # pad and slice the inputs if sp > 1
                if self.ulysses_sequence_parallel_size > 1:
                    input_ids_rmpad, position_ids_rmpad, pad_size = ulysses_pad_and_slice_inputs(input_ids_rmpad, position_ids_rmpad, sp_size=self.ulysses_sequence_parallel_size)

                # only pass input_ids and position_ids to enable flash_attn_varlen
                output = self.critic_module(
                    input_ids=input_ids_rmpad,
                    attention_mask=None,
                    position_ids=position_ids_rmpad,
                    **multi_modal_inputs,
                    use_cache=False,
                )  # prevent model thinks we are generating
                values_rmpad = output.logits
                values_rmpad = values_rmpad.squeeze(0)  # (total_nnz)

                # gather output if sp > 1
                if self.ulysses_sequence_parallel_size > 1:
                    values_rmpad = gather_outpus_and_unpad(values_rmpad, gather_dim=0, unpad_dim=0, padding_size=pad_size)

                # pad it back
                values = pad_input(values_rmpad, indices=indices, batch=batch, seqlen=seqlen).squeeze(-1)
                values = values[:, -response_length - 1 : -1]
            else:
                output = self.critic_module(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    position_ids=position_ids,
                    **multi_modal_inputs,
                    use_cache=False,
                )  # prevent model thinks we are generating
                values = output.logits
                values = values[:, -response_length - 1 : -1].squeeze(-1)
            return values

    def _optimizer_step(self):
        assert self.config.grad_clip is not None

        if isinstance(self.critic_module, FSDP):
            grad_norm = self.critic_module.clip_grad_norm_(self.config.grad_clip)
        elif isinstance(self.critic_module, FSDPModule):
            grad_norm = fsdp2_clip_grad_norm_(self.critic_module.parameters(), max_norm=self.config.grad_clip)
        else:
            grad_norm = torch.nn.utils.clip_grad_norm_(self.critic_module.parameters(), max_norm=self.config.grad_clip)

        # if grad_norm is not finite, skip the update
        if not torch.isfinite(grad_norm):
            print(f"WARN: grad_norm is not finite: {grad_norm}")
            self.critic_optimizer.zero_grad()
        else:
            self.critic_optimizer.step()
        return grad_norm

    # @GPUMemoryLogger(role="dp critic", logger=logger)
    def compute_values(self, data: DataProto) -> torch.Tensor:
        self.critic_module.eval()
        micro_batch_size = data.meta_info["micro_batch_size"]
        select_keys = ["responses", "input_ids", "attention_mask", "position_ids"]
        batch = data.select(batch_keys=select_keys).batch
        use_dynamic_bsz = data.meta_info["use_dynamic_bsz"]
        has_multi_modal_inputs = "multi_modal_inputs" in data.non_tensor_batch.keys()

        if has_multi_modal_inputs:
            num_micro_batches = data.batch.batch_size[0] // micro_batch_size
            non_tensor_select_keys = ["multi_modal_inputs"]
            micro_batches = data.select(select_keys, non_tensor_select_keys).chunk(num_micro_batches)
        elif use_dynamic_bsz:
            # split using dynamic bsz
            max_token_len = data.meta_info["max_token_len"] * self.ulysses_sequence_parallel_size
            micro_batches, indices = rearrange_micro_batches(batch=batch, max_token_len=max_token_len)
        else:
            micro_batches = batch.split(micro_batch_size)

        values_lst = []
        for micro_batch in micro_batches:
            if isinstance(micro_batch, DataProto):
                micro_batch = {**micro_batch.batch, **micro_batch.non_tensor_batch}

            with torch.no_grad():
                values = self._forward_micro_batch(micro_batch)
            values_lst.append(values)
        values = torch.concat(values_lst, dim=0)

        if use_dynamic_bsz:
            indices = list(itertools.chain.from_iterable(indices))
            assert len(indices) == values.size(0), f"{len(indices)} vs. {values.size()}"
            revert_indices = torch.tensor(get_reverse_idx(indices), dtype=torch.long)
            values = values[revert_indices]

        responses = data.batch["responses"]
        response_mask = data.batch["response_mask"]

        values = values * response_mask  # Only action tokens have values
        return values

    # @GPUMemoryLogger(role="dp critic", logger=logger)
    def update_critic(self, data: DataProto):
        # make sure we are in training mode
        self.critic_module.train()
        metrics = {}
        select_keys = ["input_ids", "responses", "attention_mask", "position_ids", "values", "returns", "response_mask"]

        batch = data.select(batch_keys=select_keys).batch
        has_multi_modal_inputs = "multi_modal_inputs" in data.non_tensor_batch.keys()

        # Split to make minibatch iterator for updating the actor
        # See PPO paper for details. https://arxiv.org/abs/1707.06347
        if has_multi_modal_inputs:
            num_mini_batches = data.batch.batch_size[0] // self.config.ppo_mini_batch_size
            non_tensor_select_keys = ["multi_modal_inputs"]
            dataloader = data.select(select_keys, non_tensor_select_keys).chunk(num_mini_batches)
        else:
            dataloader = batch.split(self.config.ppo_mini_batch_size)

        for epoch in range(self.config.ppo_epochs):
            for batch_idx, data in enumerate(dataloader):
                # split batch into micro_batches
                mini_batch = data
                if has_multi_modal_inputs:
                    num_micro_batches = mini_batch.batch.batch_size[0] // self.config.ppo_micro_batch_size_per_gpu
                    micro_batches = data.select(select_keys, non_tensor_select_keys).chunk(num_micro_batches)
                elif self.config.use_dynamic_bsz:
                    max_token_len = self.config.ppo_max_token_len_per_gpu * self.ulysses_sequence_parallel_size
                    micro_batches, _ = rearrange_micro_batches(batch=mini_batch, max_token_len=max_token_len)
                else:
                    micro_batches = mini_batch.split(self.config.ppo_micro_batch_size_per_gpu)
                self.gradient_accumulation = self.config.ppo_mini_batch_size // self.config.ppo_micro_batch_size_per_gpu

                self.critic_optimizer.zero_grad()

                for data in micro_batches:
                    # Support all devices
                    if isinstance(data, DataProto):
                        data = {**data.batch.to(get_device_id()), **data.non_tensor_batch}
                    else:
                        data = data.to(get_device_id())  # critic device is cpu when using offload
                    responses = data["responses"]
                    attention_mask = data["attention_mask"]
                    values = data["values"]
                    returns = data["returns"]
                    response_length = responses.size(1)
                    response_mask = data["response_mask"]

                    vpreds = self._forward_micro_batch(data)

                    # assert not torch.any(torch.isnan(vpreds)).item()

                    vf_loss, vf_clipfrac = core_algos.compute_value_loss(
                        vpreds=vpreds,
                        values=values,
                        returns=returns,
                        response_mask=response_mask,
                        cliprange_value=self.config.cliprange_value,
                        loss_agg_mode=self.config.loss_agg_mode,
                    )
                    if self.config.use_dynamic_bsz:
                        # relative to the dynamic bsz
                        loss = vf_loss * (len(data) / self.config.ppo_mini_batch_size)
                    else:
                        loss = vf_loss / self.gradient_accumulation

                    loss.backward()

                    data = {
                        "critic/vf_loss": vf_loss.detach().item(),
                        "critic/vf_clipfrac": vf_clipfrac.detach().item(),
                        "critic/vpred_mean": masked_mean(vpreds, response_mask).detach().item(),
                    }

                    append_to_dict(metrics, data)

                grad_norm = self._optimizer_step()
                data = {"critic/grad_norm": grad_norm.detach().item()}
                append_to_dict(metrics, data)
        self.critic_optimizer.zero_grad()
        return metrics



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import json
import logging
from typing import Any

from fastmcp import Client
from fastmcp.client.transports import SSETransport

from siirl.multiturn.tools.utils.mcp_clients.utils import TokenBucket, mcp2openai

logger = logging.getLogger(__name__)


class MCPClientManager:
    rootServerName = "mcpServers"
    initialized = False
    clients = []
    tool_client_mapping = {}
    rate_limiter = None

    async def initialize(self, config_path, rate_limit: float = 10.0):
        if self.initialized:
            return
        """Initialize the MCP Client Manager and start all clients"""
        result = self._load_config(config_path)
        servers = result[self.rootServerName]
        exclude_sse_servers = {self.rootServerName: {}}
        for server_name in servers.keys():
            server = servers[server_name]
            if "auth_token" in server:
                transport = SSETransport(url=server["url"], headers={"Authorization": f"Bearer {server['auth_token']}"})
                client = Client(transport)
                self.clients.append(client)
            else:
                exclude_sse_servers[self.rootServerName][server_name] = server

        if exclude_sse_servers[self.rootServerName]:
            self.clients.append(Client(exclude_sse_servers))

        # Initialize rate limiter
        self.rate_limiter = TokenBucket(rate_limit)
        self.initialized = True

    async def call_tool(self, tool_name, parameters, timeout):
        # Apply rate limiting
        while not self.rate_limiter.acquire():
            await asyncio.sleep(0.1)

        client = self.get_client_with_tool_name(tool_name)
        async with client:
            return await client.call_tool_mcp(tool_name, parameters)

    async def fetch_tool_schemas(self, tool_selected_list: list[str]) -> list[dict]:
        tool_schemas = []
        for client in self.clients:
            async with client:
                tools = await client.list_tools_mcp()
                for tool in tools.tools:
                    if not tool_selected_list:
                        self.tool_client_mapping[tool.name] = client
                        tool_schemas.append(mcp2openai(tool))
                    elif tool.name in tool_selected_list:
                        self.tool_client_mapping[tool.name] = client
                        tool_schemas.append(mcp2openai(tool))

        return tool_schemas

    def get_client_with_tool_name(self, tool_name: str):
        return self.tool_client_mapping[tool_name]

    def _load_config(self, file: str) -> dict[str, Any]:
        try:
            with open(file) as f:
                return json.load(f)
        except FileNotFoundError:
            logger.warning(f'the "{file}" file was not found')
        except Exception:
            logger.error(f'there was an error reading the "{file}" file')

        return {}


ClientManager = MCPClientManager()



# Copyright 2023-2024 SGLang Team
# Copyright 2025 ModelBest Inc. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# Copyright 2024 Bytedance Ltd. and/or its affiliates
# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union

from loguru import logger
import torch
import asyncio
import torch.distributed as dist
from sglang.srt.entrypoints.engine import Engine
from sglang.srt.model_executor.model_runner import LocalSerializedTensor
from sglang.srt.utils import MultiprocessingSerializer
from torch.distributed.device_mesh import DeviceMesh
from torch.distributed.fsdp.api import FullStateDictConfig, ShardedStateDictConfig, StateDictType
from torch.distributed.fsdp.fully_sharded_data_parallel import FullyShardedDataParallel as FSDP
from torch.distributed.tensor import DTensor

from siirl import DataProto
from siirl.workers.databuffer.protocol import all_gather_data_proto
from siirl.utils.debug import log_gpu_memory_usage
from siirl.utils.extras.device import get_device_id, get_device_name, get_torch_device
from siirl.utils.model_utils.fsdp_utils import fsdp_version, load_fsdp_model_to_gpu, offload_fsdp_model_to_cpu
from siirl.utils.model_utils.torch_functional import broadcast_dict_tensor, check_device_is_available
from loguru import logger
from siirl.workers.sharding_manager.base import BaseShardingManager
from siirl.utils.extras.device import get_device_id, get_torch_device
from siirl.utils.model_utils.model import convert_weight_keys
# from vllm.distributed import parallel_state as sglang_ps


def _preprocess_tensor_for_update_weights(tensor: torch.Tensor):
    if isinstance(tensor, DTensor):
        return tensor.full_tensor()
    return tensor


class MultiAgentFSDPSGLangShardingManager(BaseShardingManager):
    @check_device_is_available()
    def __init__(
        self,
        module: FSDP,
        inference_engine:  Engine,
        model_config,
        device_mesh: torch.distributed.DeviceMesh, 
        rollout_config: dict[str, int],
        full_params: bool = False,
        offload_param: bool = False,
        multi_stage_wake_up: bool = False
    ):
        self.module = module
        self.inference_engine = inference_engine
        self.model_config = model_config
        self.device_mesh = device_mesh
        self.offload_param = offload_param

        # Full params
        self.full_params = full_params
        if full_params and fsdp_version(self.module) == 1:
            FSDP.set_state_dict_type(self.module, state_dict_type=StateDictType.FULL_STATE_DICT, state_dict_config=FullStateDictConfig())
        elif fsdp_version(self.module) == 1:
            FSDP.set_state_dict_type(
                self.module,
                state_dict_type=StateDictType.SHARDED_STATE_DICT,
                state_dict_config=ShardedStateDictConfig(),
            )

        self.tp_size = self.device_mesh["infer_tp"].size()
        self.tp_rank = self.device_mesh["infer_tp"].get_local_rank()

        # Note that torch_random_states may be different on each dp rank
        self.torch_random_states = torch.cuda.get_rng_state()
        # get a random rng states
        if self.device_mesh is not None:
            gen_dp_rank = self.device_mesh["dp"].get_local_rank()
            torch.cuda.manual_seed(gen_dp_rank + 1000)  # make sure all tp ranks have the same random states
            self.gen_random_states = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.torch_random_states)
        else:
            self.gen_random_states = None

    def __enter__(self):
        torch.cuda.empty_cache()
        log_gpu_memory_usage("Before state_dict() in sharding manager memory", logger=logger)
        if self.offload_param:
            load_fsdp_model_to_gpu(self.module)
        params = self.module.state_dict()
        log_gpu_memory_usage("After state_dict() in sharding manager memory", logger=logger)
        device = torch.cuda.current_device()  # used when fsdp2 set cpu_offload_policy
        params = {k: v.to(device, non_blocking=True) if fsdp_version(self.module) == 2 else v for k, v in params.items()}
        # Copy, not share memory
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.update_weights(params))
        log_gpu_memory_usage("After sync model weights in sharding manager", logger=logger)

        del params
        if self.offload_param:
            offload_fsdp_model_to_cpu(self.module)
        torch.cuda.empty_cache()
        log_gpu_memory_usage("After del state_dict and empty_cache in sharding manager", logger=logger)

        # important: need to manually set the random states of each tp to be identical.
        if self.device_mesh is not None:
            self.torch_random_states = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.gen_random_states)

    def __exit__(self, exc_type, exc_value, traceback):
        log_gpu_memory_usage("Before SGLang offload in sharding manager", logger=logger)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.release_memory())
        log_gpu_memory_usage("After SGLang offload in sharding manager", logger=logger)

        self.module.train()

        # add empty cache after each compute
        torch.cuda.empty_cache()

        # restore random states
        if self.device_mesh is not None:
            self.gen_random_states = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.torch_random_states)

    async def update_weights(self, params):
        if self.device_mesh["infer_tp"].get_local_rank() == 0:
            await self.inference_engine.resume_memory_occupation()

        # Most naive implementation, can optimize a lot if it is bottleneck from sglang Engine weight update
        named_tensors = [(k, v) for k, v in params.items()]
        load_format = None
        for tensor_index, (name, tensor) in enumerate(named_tensors):
            serialized_tensor = MultiprocessingSerializer.serialize(_preprocess_tensor_for_update_weights(tensor))

            if self.device_mesh["infer_tp"].get_local_rank() == 0:
                gathered_serialized_tensors = [None for _ in range(self.device_mesh["infer_tp"].mesh.size()[0])]
            else:
                gathered_serialized_tensors = None
            dist.gather_object(
                obj=serialized_tensor,
                object_gather_list=gathered_serialized_tensors,
                dst=self.device_mesh["infer_tp"].mesh.tolist()[0],
                group=self.device_mesh["infer_tp"].get_group(),
            )

            if self.device_mesh["infer_tp"].get_local_rank() == 0:
                await self.inference_engine.update_weights_from_tensor(
                    named_tensors=[
                        (
                            name,
                            LocalSerializedTensor(values=gathered_serialized_tensors),
                        )
                    ],
                    load_format=load_format,
                    flush_cache=tensor_index == len(named_tensors) - 1,
                )

    async def release_memory(self):
        if self.device_mesh["infer_tp"].get_local_rank() == 0:
            await self.inference_engine.release_memory_occupation()

    async def wake_up(self):
        torch.cuda.empty_cache()
        log_gpu_memory_usage("Before state_dict() in sharding manager memory", logger=logger)
        if self.offload_param:
            load_fsdp_model_to_gpu(self.module)
        params = self.module.state_dict()
        log_gpu_memory_usage("After state_dict() in sharding manager memory", logger=logger)
        device = torch.cuda.current_device()  # used when fsdp2 set cpu_offload_policy
        params = {k: v.to(device, non_blocking=True) if fsdp_version(self.module) == 2 else v for k, v in params.items()}
        # Copy, not share memory
        await self.update_weights(params)
        log_gpu_memory_usage("After sync model weights in sharding manager", logger=logger)

        del params
        if self.offload_param:
            offload_fsdp_model_to_cpu(self.module)
        torch.cuda.empty_cache()
        log_gpu_memory_usage("After del state_dict and empty_cache in sharding manager", logger=logger)

        # important: need to manually set the random states of each tp to be identical.
        if self.device_mesh is not None:
            self.torch_random_states = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.gen_random_states)

    async def sleep(self):
        log_gpu_memory_usage("Before SGLang offload in sharding manager", logger=logger)
        await self.release_memory()
        log_gpu_memory_usage("After SGLang offload in sharding manager", logger=logger)

        self.module.train()

        # add empty cache after each compute
        torch.cuda.empty_cache()

        # restore random states
        if self.device_mesh is not None:
            self.gen_random_states = torch.cuda.get_rng_state()
            torch.cuda.set_rng_state(self.torch_random_states)

    def preprocess_data(self, data: DataProto) -> DataProto:
        """All gather across tp group to make each rank has identical input."""
        if self.tp_size == 1:
            return data

        # TODO: Current impl doesn't consider FSDP with torch micro-dp
        group = self.device_mesh["infer_tp"].get_group()

        all_gather_data_proto(data=data, process_group=group)
        return data

    def postprocess_data(self, data: DataProto) -> DataProto:
        """Get chunk data of this tp rank since we do all gather in preprocess."""
        if self.tp_size == 1:
            return data

        return data.chunk(chunks=self.tp_size)[self.tp_rank]



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from loguru import logger
import os
import sys

SIIRL_LOG_DIRCTORY = os.getenv("SIIRL_LOG_DIRECTORY", "siirl_logs")

SIIRL_LOGGING_FILENAME = os.getenv("SIIRL_LOGGING_FILENAME", "siirl")
if SIIRL_LOGGING_FILENAME != "":
    SIIRL_LOGGING_FILENAME += "_"


def set_basic_config():
    """
    This function sets the global logging format and level. It will be called when import siirl
    """
    logger.remove()
    logger.level("CRITICAL", color="<bold white on red>")
    # logger.level("ERROR", color="<red><bold>")
    # logger.level("WARNING", color="<yellow>")
    logger.add(
        sys.stderr,
        format=(
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
            "<level>{level.icon}</level> "
            "<level>{level: <8}</level> | "
            # Added :<magenta>{line}</magenta> to include the line number
            "<blue>{file}</blue>:<magenta>{line}</magenta>:<cyan>{function}</cyan> >> "
            "<level>{message}</level>"
        ),
        enqueue=True,
        colorize=True,
        level="INFO",
    )
    os.makedirs(SIIRL_LOG_DIRCTORY, exist_ok=True)

    logger.add(
        sink=os.path.join(SIIRL_LOG_DIRCTORY, SIIRL_LOGGING_FILENAME + "{time:YYYY-MM-DD-HH}.log"),
        rotation="500 MB",
        retention="30 days",
        format="{time:YYYY-MM-DD HH:mm:ss.SSS} | {level} | {module}:{function}:{line} >> {message}",
        compression="zip",
        encoding="utf-8",
        level="DEBUG",
        enqueue=True,
    )



# Copyright 2024 Bytedance Ltd. and/or its affiliates
# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
The main entry point to run the PPO algorithm
"""

import json
import os
import warnings
from dataclasses import asdict
from typing import Union, Optional

import psutil
import torch
import torch.distributed
from codetiming import Timer
from loguru import logger
from omegaconf import DictConfig
from peft import LoraConfig, TaskType, get_peft_model
from safetensors.torch import save_file
from torch.distributed import ProcessGroup, init_device_mesh
from torch.distributed.device_mesh import DeviceMesh

import siirl.utils.model_utils.torch_functional as F
from siirl import DataProto
from siirl.models.loader import load_tokenizer
from siirl.models.transformers.monkey_patch import apply_monkey_patch
from siirl.workers.base_worker import Worker
from siirl.scheduler.enums import Role
from siirl.utils.model_utils.activation_offload import enable_activation_offloading
from siirl.utils.checkpoint.fsdp_checkpoint_manager import FSDPCheckpointManager
from siirl.utils.debug import log_gpu_memory_usage
from siirl.utils.extras.device import get_device_id, get_device_name, get_nccl_backend, get_torch_device, is_cuda_available, is_npu_available
from siirl.utils.model_utils.flops_counter import FlopsCounter
from siirl.utils.extras.fs import copy_to_local
from siirl.utils.model_utils.fsdp_utils import (
    CPUOffloadPolicy,
    MixedPrecisionPolicy,
    apply_fsdp2,
    fsdp2_load_full_state_dict,
    fsdp_version,
    get_fsdp_wrap_policy,
    get_init_weight_context_manager,
    init_fn,
    layered_summon_lora_params,
    load_fsdp_model_to_gpu,
    load_fsdp_optimizer,
    offload_fsdp_model_to_cpu,
    offload_fsdp_optimizer,
)
from siirl.utils.extras.import_utils import import_external_libs
from siirl.utils.model_utils.model import compute_position_id_with_mask
from siirl.utils.extras.py_functional import convert_to_regular_types
from siirl.utils.params.model_args import ActorRolloutRefArguments, CriticArguments, FSDPArguments, OptimizerArguments, RewardModelArguments
from siirl.workers.sharding_manager.fsdp_ulysses import FSDPUlyssesShardingManager

device_name = get_device_name()


def create_device_mesh_from_group(
    process_group: ProcessGroup,
    fsdp_size: int = 1,
    sp_size: int = 1,
) -> DeviceMesh:
    """Creates a DeviceMesh from a process group for specific parallel strategies.

    This function configures a DeviceMesh based on the provided parallelism sizes.
    It supports three mutually exclusive parallelism configurations:
    1.  Data Parallelism + Sequence Parallelism ([dp, sp]): Activated when `sp_size` > 1.
    2.  Fully Sharded Data Parallelism ([fsdp]): Activated when `fsdp_size` is 1.
    3.  Distributed Data Parallelism + FSDP ([ddp, fsdp]): Activated when `fsdp_size` > 1.

    Note:
        `sp_size > 1` and `fsdp_size > 1` cannot be used simultaneously.

    Args:
        process_group (ProcessGroup): The base process group for the mesh.
        fsdp_size (int): The size of the FSDP dimension. Activates FSDP modes.
                         Defaults to 1.
        sp_size (int): The size of the Sequence Parallel dimension. Activates [dp, sp]
                       mode if > 1. Defaults to 1.

    Returns:
        DeviceMesh: A configured DeviceMesh object for the specified topology.

    Raises:
        ValueError: If inputs are invalid, parallelism strategies are mixed,
                    or the group size is not compatible with the requested
                    parallelism dimensions.
    """
    if process_group is None:
        raise ValueError("`process_group` cannot be None.")

    if sp_size > 1 and fsdp_size > 1:
        raise ValueError("Sequence Parallelism (sp_size > 1) and FSDP (fsdp_size > 1) are mutually exclusive and cannot be activated simultaneously.")

    import torch.distributed

    device_type = get_device_name()
    group_size = torch.distributed.get_world_size(group=process_group)
    ranks_in_group = torch.distributed.get_process_group_ranks(process_group)

    # --- 2. [dp, sp] Mode ---
    if sp_size > 1:
        if group_size % sp_size != 0:
            raise ValueError(f"For [dp, sp] mode, the process group size ({group_size}) must be divisible by sp_size ({sp_size}).")
        dp_size = group_size // sp_size
        mesh_shape = (dp_size, sp_size)
        mesh_dim_names = ("dp", "sp")
        logger.info(f"Creating [dp, sp] DeviceMesh with shape {mesh_shape}.")

        rank_mesh = torch.tensor(ranks_in_group, dtype=torch.long).view(mesh_shape)
        return DeviceMesh(device_type, rank_mesh, mesh_dim_names=mesh_dim_names)

    # --- 3. FSDP / DDP Modes ---
    if fsdp_size < 0 or fsdp_size >= group_size:
        # Pure FSDP (equivalent to DDP over the whole group).
        # This creates a 1D mesh representing a single shard group over all ranks.
        logger.info("Creating pure [fsdp] DeviceMesh from the process group.")
        # mesh_tensor = torch.tensor(ranks_in_group)
        return DeviceMesh.from_group(group=process_group, device_type=device_type, mesh=ranks_in_group, mesh_dim_names=("fsdp",))

    # [ddp, fsdp] mode for fsdp_size > 1
    if group_size % fsdp_size != 0:
        raise ValueError(f"The process group size ({group_size}) must be divisible by fsdp_size ({fsdp_size}).")

    # [ddp, fsdp] mode for fsdp_size > 1
    ddp_size = group_size // fsdp_size
    mesh_shape = (ddp_size, fsdp_size)
    mesh_dim_names = ("ddp", "fsdp")
    logger.info(f"Creating [ddp, fsdp] DeviceMesh with shape {mesh_shape}.")

    rank_mesh = torch.tensor(ranks_in_group, dtype=torch.long).view(mesh_shape)
    # TODO: support 2D process group(List)
    return DeviceMesh(device_type=device_type, mesh=rank_mesh, mesh_dim_names=mesh_dim_names)


def get_sharding_strategy(device_mesh):
    from torch.distributed.fsdp import ShardingStrategy

    if device_mesh.ndim == 1:
        sharding_strategy = ShardingStrategy.FULL_SHARD
    elif device_mesh.ndim == 2:
        sharding_strategy = ShardingStrategy.HYBRID_SHARD
    else:
        raise NotImplementedError(f"Get device mesh ndim={device_mesh.ndim}, but only support 1 or 2")
    return sharding_strategy


class ActorRolloutRefWorker(Worker):
    """
    This worker can be instantiated as a standalone actor or a standalone rollout or a standalone reference policy
    or a hybrid engine based on the config.rollout
    """

    def __init__(self, config: ActorRolloutRefArguments, role: str, process_group: ProcessGroup):
        super().__init__()
        self.config = config

        import torch.distributed

        if not torch.distributed.is_initialized():
            rank = int(os.environ.get("RANK", 0))
            world_size = int(os.environ.get("WORLD_SIZE", 1))
            torch.distributed.init_process_group(backend=f"cpu:gloo,{get_device_name()}:{get_nccl_backend()}", rank=rank, world_size=world_size)

        self.group_world_size = torch.distributed.get_world_size(group=process_group)
        # build device mesh for FSDP
        # TODO(sgm): support FSDP hybrid shard for larger model
        self.device_mesh = create_device_mesh_from_group(process_group=process_group, fsdp_size=self.config.actor.fsdp_config.fsdp_size)

        # build device mesh for Ulysses Sequence Parallel
        self.ulysses_device_mesh = None
        self.ulysses_sequence_parallel_size = self.config.actor.ulysses_sequence_parallel_size
        if self.ulysses_sequence_parallel_size > 1:
            self.ulysses_device_mesh = create_device_mesh_from_group(process_group=process_group, sp_size=self.ulysses_sequence_parallel_size)

        self.ulysses_sharding_manager = FSDPUlyssesShardingManager(self.ulysses_device_mesh)
        self._lora_rank = self.config.model.lora_rank
        self._is_lora = self._lora_rank > 0

        self.role = role
        assert self.role in ["actor", "rollout", "ref", "actor_rollout", "actor_rollout_ref"]

        self._is_actor = self.role in ["actor", "actor_rollout", "actor_rollout_ref"]
        self._is_rollout = self.role in ["rollout", "actor_rollout", "actor_rollout_ref"]
        self._is_ref = self.role in ["ref", "actor_rollout_ref"]

        self._is_offload_param = False
        self._is_offload_optimizer = False
        if self._is_actor:
            self._is_offload_param = self.config.actor.fsdp_config.param_offload
            self._is_offload_optimizer = self.config.actor.fsdp_config.optimizer_offload
        elif self._is_ref:
            # TODO: it seems that manual offload is slowly than FSDP offload
            self._is_offload_param = self.config.ref.fsdp_config.param_offload

        # normalize config
        if self._is_actor:
            self.config.actor.ppo_mini_batch_size *= self.config.rollout.n
            self.config.actor.ppo_mini_batch_size //= self.device_mesh.size() // self.ulysses_sequence_parallel_size
            assert self.config.actor.ppo_mini_batch_size > 0, f"ppo_mini_batch_size {self.config.actor.ppo_mini_batch_size} should be larger than 0 after normalization"
            # micro bsz
            if self.config.actor.ppo_micro_batch_size is not None:
                self.config.actor.ppo_micro_batch_size //= self.device_mesh.size() // self.ulysses_sequence_parallel_size
                self.config.actor.ppo_micro_batch_size_per_gpu = self.config.actor.ppo_micro_batch_size

            if self.config.actor.ppo_micro_batch_size_per_gpu is not None:
                assert self.config.actor.ppo_mini_batch_size % self.config.actor.ppo_micro_batch_size_per_gpu == 0, f"normalized ppo_mini_batch_size {self.config.actor.ppo_mini_batch_size} should be divisible by ppo_micro_batch_size_per_gpu {self.config.actor.ppo_micro_batch_size_per_gpu}"
                assert self.config.actor.ppo_mini_batch_size // self.config.actor.ppo_micro_batch_size_per_gpu > 0, f"normalized ppo_mini_batch_size {self.config.actor.ppo_mini_batch_size} should be larger than ppo_micro_batch_size_per_gpu {self.config.actor.ppo_micro_batch_size_per_gpu}"

        # normalize rollout config
        if self._is_rollout and self.config.rollout.log_prob_micro_batch_size is not None:
            self.config.rollout.log_prob_micro_batch_size //= self.device_mesh.size() // self.ulysses_sequence_parallel_size
            self.config.rollout.log_prob_micro_batch_size_per_gpu = self.config.rollout.log_prob_micro_batch_size
        # normalize ref config
        if self._is_ref and self.config.ref.log_prob_micro_batch_size is not None:
            self.config.ref.log_prob_micro_batch_size //= self.device_mesh.size() // self.ulysses_sequence_parallel_size
            self.config.ref.log_prob_micro_batch_size_per_gpu = self.config.ref.log_prob_micro_batch_size

    def _build_model_optimizer(
        self,
        model_path: str,
        fsdp_config: FSDPArguments,
        optim_config: Optional[OptimizerArguments],
        override_model_config: DictConfig,
        use_remove_padding: bool = False,
        use_fused_kernels: bool = False,
        enable_gradient_checkpointing: bool = False,
        trust_remote_code: bool = False,
        use_liger: bool = False,
        role: Role = Role.Actor,
        enable_activation_offload: bool = False,
    ):
        from torch import optim
        from torch.distributed.fsdp import CPUOffload, MixedPrecision
        from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
        from transformers import AutoConfig, AutoModelForCausalLM, AutoModelForVision2Seq

        from siirl.utils.model_utils.model import get_generation_config, print_model_size, update_model_config
        from siirl.utils.model_utils.torch_dtypes import PrecisionType

        assert role in [Role.Actor, Role.RefPolicy]

        log_gpu_memory_usage(f"Before init {role} from HF AutoModel", logger=logger)
        local_path = model_path

        torch_dtype = fsdp_config.model_dtype
        if torch_dtype is None:
            torch_dtype = torch.float32 if self._is_actor else torch.bfloat16
        else:
            torch_dtype = PrecisionType.to_dtype(torch_dtype)

        # override model kwargs
        actor_model_config = AutoConfig.from_pretrained(local_path, trust_remote_code=trust_remote_code, attn_implementation="flash_attention_2")
        if self._is_ref:
            self.flops_counter = FlopsCounter(actor_model_config, forward_only=True)
        # patch for kimi-vl
        if getattr(actor_model_config, "model_type", None) == "kimi_vl":
            actor_model_config.text_config.topk_method = "greedy"

        self.generation_config = get_generation_config(local_path, trust_remote_code=trust_remote_code)

        override_config_kwargs = {
            "bos_token_id": self.tokenizer.bos_token_id,
            "eos_token_id": self.tokenizer.eos_token_id,
            "pad_token_id": self.tokenizer.pad_token_id,
        }
        override_config_kwargs.update(override_model_config)
        update_model_config(actor_model_config, override_config_kwargs=override_config_kwargs)
        if self.rank == 0:
            logger.info(f"Model config after override: {actor_model_config}")

        # NOTE(fix me): tie_word_embedding causes meta_tensor init to hang
        init_context = get_init_weight_context_manager(use_meta_tensor=not actor_model_config.tie_word_embeddings, mesh=self.device_mesh)

        with init_context(), warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if type(actor_model_config) in AutoModelForVision2Seq._model_mapping.keys():
                actor_module_class = AutoModelForVision2Seq
            elif type(actor_model_config).__name__ in "configuration_internvl_chat.InternVLChatConfig":
                from siirl.models.transformers.internvl_chat import InternVLChatModel

                actor_module_class = InternVLChatModel
                logger.info("Choose InternVLChatModel for internvl")
            else:
                actor_module_class = AutoModelForCausalLM

            actor_module = actor_module_class.from_pretrained(
                pretrained_model_name_or_path=local_path,
                torch_dtype=torch_dtype,
                config=actor_model_config,
                trust_remote_code=trust_remote_code,
            )

            if type(actor_model_config).__name__ in "configuration_internvl_chat.InternVLChatConfig":
                from siirl.models.transformers.internvl import IMG_CONTEXT_TOKEN

                actor_module.img_context_token_id = self.tokenizer.convert_tokens_to_ids(IMG_CONTEXT_TOKEN)

            # Apply Liger kernel to the model if use_liger is set to True
            if use_liger:
                from liger_kernel.transformers.monkey_patch import _apply_liger_kernel_to_instance

                _apply_liger_kernel_to_instance(model=actor_module)

            apply_monkey_patch(
                model=actor_module,
                use_remove_padding=use_remove_padding,
                ulysses_sp_size=self.ulysses_sequence_parallel_size,
                use_fused_kernels=use_fused_kernels,
            )

            # some parameters may not in torch_dtype. TODO(zhangchi.usc1992) remove this after we switch to fsdp2
            actor_module.to(torch_dtype)

            if enable_gradient_checkpointing:
                actor_module.gradient_checkpointing_enable(gradient_checkpointing_kwargs={"use_reentrant": False})
            if self._is_lora:
                logger.info("Applying LoRA to actor module")
                actor_module.enable_input_require_grads()
                # Convert config to regular Python types before creating PEFT model
                lora_config = {"task_type": TaskType.CAUSAL_LM, "r": self.config.model.lora_rank, "lora_alpha": self.config.model.lora_alpha, "target_modules": convert_to_regular_types(self.config.model.target_modules), "bias": "none"}
                actor_module = get_peft_model(actor_module, LoraConfig(**lora_config))
        torch.distributed.barrier()

        if self.rank == 0:
            print_model_size(actor_module)

        log_gpu_memory_usage(f"After init {role} from HF AutoModel", logger=logger)

        # We wrap FSDP for rollout as well
        mixed_precision_config = fsdp_config.mixed_precision
        if mixed_precision_config is not None:
            param_dtype = PrecisionType.to_dtype(mixed_precision_config.param_dtype)
            reduce_dtype = PrecisionType.to_dtype(mixed_precision_config.reduce_dtype)
            buffer_dtype = PrecisionType.to_dtype(mixed_precision_config.buffer_dtype)
        else:
            param_dtype = torch.bfloat16
            reduce_dtype = torch.float32
            buffer_dtype = torch.float32

        mixed_precision = MixedPrecision(param_dtype=param_dtype, reduce_dtype=reduce_dtype, buffer_dtype=buffer_dtype)

        auto_wrap_policy = get_fsdp_wrap_policy(module=actor_module, config=fsdp_config.wrap_policy, is_lora=self.config.model.lora_rank > 0)

        if self._is_rollout and self.config.rollout.name == "hf":
            # TODO(zhangchi.usc1992, shengguangming) fix me. Current, auto_wrap_policy causes HFRollout to hang in Gemma
            auto_wrap_policy = None

        if self.rank == 0:
            logger.info(f"wrap_policy: {auto_wrap_policy}")

        fsdp_mesh = self.device_mesh
        sharding_strategy = get_sharding_strategy(fsdp_mesh)

        # TODO: add transformer policy
        # We force reference policy to use CPUOffload to save memory.
        # We force turn off CPUOffload for actor because it causes incorrect results when using grad accumulation
        cpu_offload = None if role == Role.Actor else CPUOffload(offload_params=True)
        fsdp_strategy = self.config.actor.strategy
        if fsdp_strategy == "fsdp":
            actor_module_fsdp = FSDP(
                actor_module,
                cpu_offload=cpu_offload,
                param_init_fn=init_fn,
                use_orig_params=False,
                auto_wrap_policy=auto_wrap_policy,
                device_id=get_device_id(),
                sharding_strategy=sharding_strategy,  # zero3
                mixed_precision=mixed_precision,
                sync_module_states=True,
                device_mesh=self.device_mesh,
                forward_prefetch=False,
            )
        elif fsdp_strategy == "fsdp2":
            assert CPUOffloadPolicy is not None, "PyTorch version >= 2.4 is required for using fully_shard API (FSDP2)"
            mp_policy = MixedPrecisionPolicy(param_dtype=param_dtype, reduce_dtype=reduce_dtype, cast_forward_inputs=True)
            if role == Role.Actor and fsdp_config.offload_policy:
                cpu_offload = CPUOffloadPolicy(pin_memory=True)
                self._is_offload_param = False
                self._is_offload_optimizer = False
            else:
                cpu_offload = None if role == Role.Actor else CPUOffloadPolicy(pin_memory=True)

            fsdp_kwargs = {
                "mesh": fsdp_mesh,
                "mp_policy": mp_policy,
                "offload_policy": cpu_offload,
                "reshard_after_forward": fsdp_config.reshard_after_forward,
            }
            full_state = actor_module.state_dict()
            apply_fsdp2(actor_module, fsdp_kwargs, fsdp_config)
            fsdp2_load_full_state_dict(actor_module, full_state, fsdp_mesh, cpu_offload)
            actor_module_fsdp = actor_module
        else:
            raise NotImplementedError(f"not implement {fsdp_strategy}")

        if enable_activation_offload:
            enable_activation_offloading(actor_module_fsdp, fsdp_strategy, enable_gradient_checkpointing)

        log_gpu_memory_usage(f"After {role} FSDP init", logger=logger)

        # TODO: add more optimizer args into config
        if role == Role.Actor and optim_config is not None:
            from siirl.utils.model_utils.torch_functional import get_constant_schedule_with_warmup, get_cosine_schedule_with_warmup

            actor_optimizer = optim.AdamW(
                actor_module_fsdp.parameters(),
                lr=optim_config.lr,
                betas=optim_config.betas,
                weight_decay=optim_config.weight_decay,
            )

            total_steps = optim_config.total_training_steps
            num_warmup_steps = int(optim_config.lr_warmup_steps)
            warmup_style = optim_config.warmup_style
            min_lr_ratio = optim_config.min_lr_ratio if optim_config.min_lr_ratio else 0.0
            num_cycles = optim_config.num_cycles
            if num_warmup_steps < 0:
                num_warmup_steps_ratio = optim_config.lr_warmup_steps_ratio
                num_warmup_steps = int(num_warmup_steps_ratio * total_steps)

            if self.rank == 0:
                logger.info(f"Total steps: {total_steps}, num_warmup_steps: {num_warmup_steps}")

            if warmup_style == "constant":
                actor_lr_scheduler = get_constant_schedule_with_warmup(optimizer=actor_optimizer, num_warmup_steps=num_warmup_steps)
            elif warmup_style == "cosine":
                actor_lr_scheduler = get_cosine_schedule_with_warmup(optimizer=actor_optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=total_steps, min_lr_ratio=min_lr_ratio, num_cycles=num_cycles)
            else:
                raise NotImplementedError(f"Warmup style {warmup_style} is not supported")

            log_gpu_memory_usage(f"After {role} optimizer init", logger=logger)
        else:
            actor_optimizer = None
            actor_lr_scheduler = None

        return actor_module_fsdp, actor_optimizer, actor_lr_scheduler, actor_model_config

    def _build_rollout(self, trust_remote_code=False):
        from siirl.utils.model_utils.model import get_generation_config

        local_path = copy_to_local(self.config.model.path)
        self.generation_config = get_generation_config(local_path, trust_remote_code=trust_remote_code)

        # TODO(sgm): support FSDP hybrid shard for larger model
        rollout_name = self.config.rollout.name
        if rollout_name == "hf":
            from siirl.workers.rollout import HFRollout
            from siirl.workers.sharding_manager.base import BaseShardingManager

            rollout = HFRollout(module=self.actor_module_fsdp, config=self.config.rollout)
            rollout_sharding_manager = BaseShardingManager()
            # TODO: a sharding manager that do nothing?

        elif rollout_name == "vllm":
            from siirl.workers.rollout.vllm_rollout import vllm_mode, vLLMRollout

            log_gpu_memory_usage(f"Before building {rollout_name} rollout", logger=logger)
            local_path = copy_to_local(self.config.model.path, use_shm=self.config.model.use_shm)
            lora_kwargs = {"lora_kwargs": {"enable_lora": True, "max_loras": 1, "max_lora_rank": self._lora_rank}} if self._is_lora else {}
            # lora_kwargs = {}
            if vllm_mode == "customized":
                rollout = vLLMRollout(actor_module=self.actor_module_fsdp, config=self.config.rollout, tokenizer=self.tokenizer, model_hf_config=self.actor_model_config, trust_remote_code=trust_remote_code, **lora_kwargs)
            elif vllm_mode == "spmd":
                from siirl.workers.rollout.vllm_rollout import vLLMAsyncRollout

                vllm_rollout_cls = vLLMRollout if self.config.rollout.mode == "sync" else vLLMAsyncRollout
                rollout = vllm_rollout_cls(model_path=local_path, config=self.config.rollout, tokenizer=self.tokenizer, model_hf_config=self.actor_model_config, trust_remote_code=trust_remote_code, **lora_kwargs)
            else:
                raise NotImplementedError("vllm_mode must be 'customized' or 'spmd'")

            log_gpu_memory_usage(f"After building {rollout_name} rollout", logger=logger)
            if self.device_mesh.mesh.numel() == 1:
                self.config.rollout.load_format = "dummy_hf"
            log_gpu_memory_usage("After building sharding manager", logger=logger)

        elif rollout_name == "sglang":
            from siirl.workers.rollout.sglang_rollout import SGLangRollout

            # NOTE(linjunrong): Due to recent fp8 support in SGLang. Now importing any symbol relate to
            # SGLang's model_runner would check CUDA device capability. However, due to siirl's setting,
            # the main process of ray can not find any CUDA device, which would potentially lead to:
            # "RuntimeError: No CUDA GPUs are available".
            # For this reason, sharding_manager.__init__ should not import FSDPSGLangShardingManager and
            # we import it here use the abs path.
            # check: https://github.com/sgl-project/sglang/blob/00f42707eaddfc2c0528e5b1e0094025c640b7a0/python/sglang/srt/layers/quantization/fp8_utils.py#L76
            # from siirl.workers.sharding_manager.fsdp_sglang import MultiAgentFSDPSGLangShardingManager

            local_path = copy_to_local(self.config.model.path)
            log_gpu_memory_usage(f"Before building {rollout_name} rollout", logger=logger)
            rollout = SGLangRollout(
                actor_module=local_path,
                config=self.config.rollout,
                tokenizer=self.tokenizer,
                model_hf_config=self.actor_model_config,
                processing_class=self.processor if self.processor is not None else self.tokenizer,
                trust_remote_code=trust_remote_code,
            )
            log_gpu_memory_usage(f"After building {rollout_name} rollout", logger=logger)

        else:
            raise NotImplementedError(f"Rollout name: {self.config.rollout.name} is not supported")

        return rollout, None

    def init_model(self):
        from siirl.workers.actor import DataParallelPPOActor

        # This is used to import external_lib into the huggingface systems
        import_external_libs(self.config.model.external_lib)

        override_model_config = self.config.model.override_config
        use_remove_padding = self.config.model.use_remove_padding
        use_fused_kernels = self.config.model.use_fused_kernels
        use_shm = self.config.model.use_shm

        # note that we have to create model in fp32. Otherwise, the optimizer is in bf16, which is incorrect
        # TODO(zhangchi.usc1992): 1. support create from random initialized model. 2. Support init with FSDP directly
        tokenizer_module = load_tokenizer(model_args=self.config.model)
        self.tokenizer, self.processor = tokenizer_module["tokenizer"], tokenizer_module["processor"]

        if self._is_actor:
            optim_config = self.config.actor.optim
            fsdp_config = self.config.actor.fsdp_config

            local_path = copy_to_local(self.config.model.path, use_shm=use_shm)
            (
                self.actor_module_fsdp,
                self.actor_optimizer,
                self.actor_lr_scheduler,
                self.actor_model_config,
            ) = self._build_model_optimizer(
                model_path=local_path,
                fsdp_config=fsdp_config,
                optim_config=optim_config,
                override_model_config=override_model_config,
                use_remove_padding=use_remove_padding,
                use_fused_kernels=use_fused_kernels,
                enable_gradient_checkpointing=self.config.model.enable_gradient_checkpointing,
                trust_remote_code=self.config.model.trust_remote_code,
                use_liger=self.config.model.use_liger,
                role=Role.Actor,
                enable_activation_offload=self.config.model.enable_activation_offload,
            )

            # get the original unwrapped module
            if fsdp_version(self.actor_module_fsdp) == 1:
                self.actor_module = self.actor_module_fsdp._fsdp_wrapped_module

            if self._is_offload_param:
                offload_fsdp_model_to_cpu(self.actor_module_fsdp)
                log_gpu_memory_usage("After offload actor model during init", logger=logger)

            if self._is_offload_optimizer:
                offload_fsdp_optimizer(optimizer=self.actor_optimizer)
                log_gpu_memory_usage("After offload actor optimizer during init", logger=logger)
        # load from checkpoint
        if self._is_actor:
            self.config.actor.use_remove_padding = use_remove_padding
            self.config.actor.use_fused_kernels = use_fused_kernels
            self.actor = DataParallelPPOActor(config=self.config.actor, actor_module=self.actor_module_fsdp, actor_optimizer=self.actor_optimizer)

        if self._is_rollout:
            from transformers import AutoConfig

            local_path = copy_to_local(self.config.model.path)
            self.actor_model_config = AutoConfig.from_pretrained(local_path, trust_remote_code=self.config.model.trust_remote_code)
            self.flops_counter = FlopsCounter(self.actor_model_config, forward_only=True)
            self.rollout, self.rollout_sharding_manager = self._build_rollout(trust_remote_code=self.config.model.trust_remote_code)

        if self._is_ref:
            local_path = copy_to_local(self.config.model.path, use_shm=use_shm)
            self.ref_module_fsdp = self._build_model_optimizer(
                model_path=local_path,
                fsdp_config=self.config.ref.fsdp_config,
                optim_config=None,
                override_model_config=override_model_config,
                use_remove_padding=use_remove_padding,
                use_fused_kernels=use_fused_kernels,
                trust_remote_code=self.config.model.trust_remote_code,
                use_liger=self.config.model.use_liger,
                role=Role.RefPolicy,
            )[0]
            self.config.ref.use_remove_padding = use_remove_padding
            self.config.ref.use_fused_kernels = use_fused_kernels
            self.ref_policy = DataParallelPPOActor(config=self.config.ref, actor_module=self.ref_module_fsdp)

        if self._is_actor:
            self.flops_counter = FlopsCounter(self.actor_model_config)
            self.checkpoint_manager = FSDPCheckpointManager(
                model=self.actor_module_fsdp, optimizer=self.actor.actor_optimizer, lr_scheduler=self.actor_lr_scheduler, processing_class=self.processor if self.processor is not None else self.tokenizer, checkpoint_contents=self.config.actor.checkpoint.contents, tokenizer=self.tokenizer
            )

    def update_actor(self, data: DataProto):
        # Support all hardwares
        data = data.to(get_device_id())

        assert self._is_actor
        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.actor_module_fsdp)
        if self._is_offload_optimizer:
            load_fsdp_optimizer(optimizer=self.actor_optimizer, device_id=get_device_id())

        with self.ulysses_sharding_manager:
            data = self.ulysses_sharding_manager.preprocess_data(data=data)
            # perform training
            with Timer(name="update_policy", logger=None) as timer:
                metrics = self.actor.update_policy(data=data)
            delta_time = timer.last
            global_num_tokens = data.meta_info["global_token_num"]
            estimated_flops, promised_flops = self.flops_counter.estimate_flops(global_num_tokens, delta_time)
            metrics["perf/mfu/actor"] = estimated_flops * self.config.actor.ppo_epochs / promised_flops
            metrics["perf/delta_time/actor"] = delta_time
            metrics["perf/max_memory_allocated_gb"] = get_torch_device().max_memory_allocated() / (1024**3)
            metrics["perf/max_memory_reserved_gb"] = get_torch_device().max_memory_reserved() / (1024**3)
            metrics["perf/cpu_memory_used_gb"] = psutil.virtual_memory().used / (1024**3)

            lr = self.actor_lr_scheduler.get_last_lr()[0]
            metrics["actor/lr"] = lr
            self.actor_lr_scheduler.step()

            # TODO: here, we should return all metrics
            data.meta_info["metrics"] = metrics
            processed_data = self.ulysses_sharding_manager.postprocess_data(data=data)
            processed_data = processed_data.to("cpu")

        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.actor_module_fsdp)
            log_gpu_memory_usage("After offload actor model during update_actor", logger=logger)
        if self._is_offload_optimizer:
            offload_fsdp_optimizer(optimizer=self.actor_optimizer)
            log_gpu_memory_usage("After offload actor optimizer during update_actor", logger=logger)

        return processed_data

    def generate_sequences(self, prompts: DataProto):
        # Support all hardwares
        prompts = prompts.to(get_device_id())

        assert self._is_rollout

        meta_info = {
            "eos_token_id": self.generation_config.eos_token_id if self.generation_config is not None else self.tokenizer.eos_token_id,
            "pad_token_id": self.generation_config.pad_token_id if self.generation_config is not None else self.tokenizer.pad_token_id,
        }
        prompts.meta_info.update(meta_info)
        with self.rollout_sharding_manager:
            log_gpu_memory_usage("After entering rollout sharding manager", logger=logger)

            if self.config.rollout.name == "sglang_async":
                from siirl.workers.rollout.sglang_rollout import AsyncSGLangRollout

                if isinstance(self.rollout, AsyncSGLangRollout) and hasattr(self.rollout, "_tool_schemas") and len(self.rollout._tool_schemas) > 0:
                    output = self.rollout.generate_sequences_with_tools(prompts=prompts)
                else:
                    output = self.rollout.generate_sequences(prompts=prompts)
            else:
                with Timer(name="generate_sequences", logger=None) as timer:
                    output = self.rollout.generate_sequences(prompts=prompts)
                total_input_tokens = output.meta_info.get("total_input_tokens", 0)
                total_output_tokens = output.meta_info.get("total_output_tokens", 0)
                delta_time = timer.last
                estimated_flops, promised_flops = self.flops_counter.estimate_flops([total_input_tokens, total_output_tokens], delta_time)
                metrics = {}
                metrics["perf/mfu/rollout"] = estimated_flops / promised_flops / self.config.rollout.tensor_model_parallel_size
                metrics["perf/delta_time/rollout"] = delta_time
                output.meta_info.update({"metrics": metrics})
            log_gpu_memory_usage("After rollout generation", logger=logger)

        output = output.to("cpu")

        # clear kv cache
        get_torch_device().empty_cache()
        return output

    def compute_log_prob(self, data: DataProto):
        # when is_lora is True, we use the actor without lora applied to calculate the log_prob
        # which is mostly used for ref log_prob calculation
        assert self._is_actor
        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.actor_module_fsdp)

        # Support all hardwares
        from contextlib import nullcontext

        is_lora = data.meta_info.pop("is_lora", False)
        adapter_ctx = self.actor.actor_module.disable_adapter() if is_lora else nullcontext()
        data = data.to(get_device_id())
        # we should always recompute old_log_probs when it is HybridEngine
        data.meta_info["micro_batch_size"] = self.config.rollout.log_prob_micro_batch_size_per_gpu
        data.meta_info["max_token_len"] = self.config.rollout.log_prob_max_token_len_per_gpu
        data.meta_info["use_dynamic_bsz"] = self.config.rollout.log_prob_use_dynamic_bsz
        data.meta_info["temperature"] = self.config.rollout.temperature
        # perform recompute log_prob
        with self.ulysses_sharding_manager:
            data = self.ulysses_sharding_manager.preprocess_data(data)
            with Timer(name="compute_actor_log_prob", logger=None) as timer, adapter_ctx:
                output, entropys = self.actor.compute_log_prob(data=data, calculate_entropy=True)
            delta_time = timer.last
            global_num_tokens = data.meta_info["global_token_num"]
            estimated_flops, promised_flops = self.flops_counter.estimate_flops(global_num_tokens, delta_time)
            metrics = {
                # actor forward
                "perf/mfu/actor_log_prob": estimated_flops / promised_flops / 3,
                "perf/delta_time/actor_log_prob": delta_time,
            }
            data.batch["old_log_probs"] = output
            data.batch["entropys"] = entropys
            data.meta_info["metrics"] = metrics
            processed_data = self.ulysses_sharding_manager.postprocess_data(data)

        processed_data = processed_data.to("cpu")

        # https://pytorch.org/docs/stable/notes/fsdp.html#fsdp-notes
        # unshard the root FSDP module
        if self.group_world_size > 1 and fsdp_version(self.actor.actor_module) == 1:
            self.actor.actor_module._handle.reshard(True)

        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.actor_module_fsdp)
            log_gpu_memory_usage("After offload actor model during compute_log_prob", logger=logger)

        return processed_data

    def compute_ref_log_prob(self, data: DataProto):
        if self._is_lora:
            # if _is_lora, actor without lora applied is the ref
            data.meta_info["is_lora"] = True
            data = self.compute_log_prob(data)
            # this old_log_probs is in fact ref_log_prob
            data = DataProto.from_dict(tensors={"ref_log_prob": data.batch["old_log_probs"]})
            return data
        assert self._is_ref
        # else:
        # otherwise, the class have a standalone ref model
        # Support all hardwares
        data = data.to(get_device_id())

        metrics = {}
        micro_batch_size = self.config.ref.log_prob_micro_batch_size_per_gpu
        data.meta_info["micro_batch_size"] = micro_batch_size
        data.meta_info["temperature"] = self.config.rollout.temperature
        data.meta_info["max_token_len"] = self.config.ref.log_prob_max_token_len_per_gpu
        data.meta_info["use_dynamic_bsz"] = self.config.ref.log_prob_use_dynamic_bsz
        with self.ulysses_sharding_manager:
            data = self.ulysses_sharding_manager.preprocess_data(data)
            with Timer(name="compute_log_prob", logger=None) as timer:
                output, _ = self.ref_policy.compute_log_prob(data=data, calculate_entropy=False)
            delta_time = timer.last
            global_num_tokens = data.meta_info["global_token_num"]
            estimated_flops, promised_flops = self.flops_counter.estimate_flops(global_num_tokens, delta_time)
            metrics["perf/mfu/ref"] = estimated_flops / promised_flops
            metrics["perf/delta_time/ref"] = delta_time
            data.batch["ref_log_prob"] = output
            data.meta_info["metrics"] = metrics
            processed_data = self.ulysses_sharding_manager.postprocess_data(data)

        processed_data = processed_data.to("cpu")

        # https://pytorch.org/docs/stable/notes/fsdp.html#fsdp-notes
        # unshard the root FSDP module
        if self.group_world_size > 1 and fsdp_version(self.ref_policy.actor_module) == 1:
            self.ref_policy.actor_module._handle.reshard(True)

        return processed_data

    def save_checkpoint(self, local_path, hdfs_path=None, global_step=0, max_ckpt_to_keep=None):
        # only support save and load ckpt for actor
        assert self._is_actor
        import torch

        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.actor_module_fsdp)

        self.checkpoint_manager.save_checkpoint(local_path=local_path, hdfs_path=hdfs_path, global_step=global_step, max_ckpt_to_keep=max_ckpt_to_keep)

        torch.distributed.barrier()

        if self._is_lora and hasattr(getattr(self, "actor_module", self.actor_module_fsdp), "peft_config"):
            lora_save_path = os.path.join(local_path, "lora_adapter")
            peft_model = getattr(self, "actor_module", self.actor_module_fsdp)
            peft_config = {}
            if torch.distributed.get_rank() == 0:
                os.makedirs(lora_save_path, exist_ok=True)
                peft_config = asdict(peft_model.peft_config.get("default", {}))
                peft_config["task_type"] = peft_config["task_type"].value
                peft_config["peft_type"] = peft_config["peft_type"].value
                peft_config["target_modules"] = list(peft_config["target_modules"])
            try:
                if fsdp_version(self.actor_module_fsdp) > 0:
                    self.actor_module_fsdp = self.actor_module_fsdp.to(get_device_name())
                    lora_params = layered_summon_lora_params(self.actor_module_fsdp)
                    if torch.distributed.get_rank() == 0:
                        save_file(lora_params, os.path.join(lora_save_path, "adapter_model.safetensors"))
                        with open(os.path.join(lora_save_path, "adapter_config.json"), "w", encoding="utf-8") as f:
                            json.dump(peft_config, f, ensure_ascii=False, indent=4)
            except Exception as e:
                if torch.distributed.get_rank() == 0:
                    logger.info(f"[rank-{self.rank}]: Save LoRA Adapter Error ({e})")

            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                logger.info(f"[rank-{self.rank}]: Saved LoRA adapter to: {lora_save_path}")

        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.actor_module_fsdp)

    def load_checkpoint(self, local_path, hdfs_path=None, del_local_after_load=False):
        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.actor_module_fsdp)

        self.checkpoint_manager.load_checkpoint(local_path=local_path, hdfs_path=hdfs_path, del_local_after_load=del_local_after_load)

        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.actor_module_fsdp)

        if self._is_offload_optimizer:
            offload_fsdp_optimizer(self.actor_optimizer)

    def set_rollout_sharding_manager(self, sharding_manager):
        self.rollout_sharding_manager = sharding_manager


class CriticWorker(Worker):
    def __init__(self, config: CriticArguments, process_group: ProcessGroup):
        super().__init__()
        import torch.distributed

        if not torch.distributed.is_initialized():
            torch.distributed.init_process_group(backend=get_nccl_backend())
        self.config = config
        world_size = torch.distributed.get_world_size(group=process_group)
        self.group_world_size = world_size
        # build device mesh for Ulysses Sequence Parallel
        self.device_mesh = create_device_mesh_from_group(process_group=process_group, fsdp_size=self.config.model.fsdp_config.fsdp_size)

        self.ulysses_device_mesh = None
        self.ulysses_sequence_parallel_size = self.config.ulysses_sequence_parallel_size
        if self.ulysses_sequence_parallel_size > 1:
            self.ulysses_device_mesh = create_device_mesh_from_group(process_group=process_group, sp_size=self.ulysses_sequence_parallel_size)

        self.ulysses_sharding_manager = FSDPUlyssesShardingManager(self.ulysses_device_mesh)

        # set FSDP offload params
        self._is_offload_param = self.config.model.fsdp_config.param_offload
        self._is_offload_optimizer = self.config.model.fsdp_config.optimizer_offload

        # normalize config
        self.config.ppo_mini_batch_size *= self.config.rollout_n
        self.config.ppo_mini_batch_size //= world_size // self.ulysses_sequence_parallel_size
        if self.config.ppo_micro_batch_size is not None:
            self.config.ppo_micro_batch_size //= world_size // self.ulysses_sequence_parallel_size
            self.config.forward_micro_batch_size //= world_size // self.ulysses_sequence_parallel_size
            self.config.ppo_micro_batch_size_per_gpu = self.config.ppo_micro_batch_size
            self.config.forward_micro_batch_size_per_gpu = self.config.forward_micro_batch_size

        if self.config.ppo_micro_batch_size_per_gpu is not None:
            assert self.config.ppo_mini_batch_size % self.config.ppo_micro_batch_size_per_gpu == 0, f"normalized ppo_mini_batch_size {self.config.ppo_mini_batch_size} should be divisible by ppo_micro_batch_size_per_gpu {self.config.ppo_micro_batch_size_per_gpu}"
            assert self.config.ppo_mini_batch_size // self.config.ppo_micro_batch_size_per_gpu > 0, f"normalized ppo_mini_batch_size {self.config.ppo_mini_batch_size} should be larger than ppo_micro_batch_size_per_gpu {self.config.ppo_micro_batch_size_per_gpu}"
        self._is_lora = self.config.model.lora_rank > 0

    def _build_critic_model_optimizer(self, config):
        # the following line is necessary
        from torch import optim
        from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
        from torch.distributed.fsdp import MixedPrecision

        from siirl.utils.model_utils.model import print_model_size
        from siirl.utils.model_utils.torch_dtypes import PrecisionType

        use_shm = config.model.use_shm
        local_path = copy_to_local(config.model.path, use_shm=use_shm)
        # note that the tokenizer between actor and critic may be different. So override tokenizer info with actor info
        # using random initialized model from any architecture. May not be the same as Actor.

        tokenizer_module = load_tokenizer(model_args=self.config.model)
        self.tokenizer, self.processor = tokenizer_module["tokenizer"], tokenizer_module["processor"]

        override_config = self.config.model.override_config
        override_config_kwargs = {
            "bos_token_id": self.tokenizer.bos_token_id,
            "eos_token_id": self.tokenizer.eos_token_id,
            "pad_token_id": self.tokenizer.pad_token_id,
        }
        override_config_kwargs.update(override_config)
        if self.rank == 0:
            logger.info(f"Critic overriding config {override_config_kwargs}")

        torch_dtype = self.config.model.fsdp_config.model_dtype
        torch_dtype = PrecisionType.to_dtype(torch_dtype)

        from transformers import AutoConfig, AutoModelForTokenClassification

        critic_model_config = AutoConfig.from_pretrained(local_path, attn_implementation="flash_attention_2", trust_remote_code=config.model.trust_remote_code)
        critic_model_config.num_labels = 1
        # patch for kimi-vl
        if getattr(critic_model_config, "model_type", None) == "kimi_vl":
            critic_model_config.text_config.topk_method = "greedy"

        init_context = get_init_weight_context_manager(use_meta_tensor=not critic_model_config.tie_word_embeddings, mesh=self.device_mesh)

        with init_context(), warnings.catch_warnings():
            warnings.simplefilter("ignore")
            critic_model_config.classifier_dropout = 0.0
            critic_model_config.hidden_dropout = "0"
            critic_module = AutoModelForTokenClassification.from_pretrained(
                pretrained_model_name_or_path=local_path,
                torch_dtype=torch_dtype,
                config=critic_model_config,
                trust_remote_code=config.model.trust_remote_code,
            )

            use_remove_padding = config.model.use_remove_padding

            apply_monkey_patch(
                model=critic_module,
                use_remove_padding=use_remove_padding,
                ulysses_sp_size=self.ulysses_sequence_parallel_size,
            )

            # some parameters may not in torch_dtype
            critic_module.to(torch_dtype)

            if config.model.enable_gradient_checkpointing:
                critic_module.gradient_checkpointing_enable(gradient_checkpointing_kwargs={"use_reentrant": False})

        if self._is_lora:
            logger.info("Applying LoRA to critic module")
            critic_module.enable_input_require_grads()
            # Convert config to regular Python types before creating PEFT model
            lora_config = {
                "task_type": TaskType.CAUSAL_LM,
                "r": self.config.model.lora_rank,
                "lora_alpha": self.config.model.lora_alpha,
                "target_modules": convert_to_regular_types(self.config.model.target_modules),
                "bias": "none",
            }
            critic_module = get_peft_model(critic_module, LoraConfig(**lora_config))

        if self.rank == 0:
            print_model_size(critic_module)

        self.critic_model_config = critic_model_config

        fsdp_config = self.config.model.fsdp_config
        mixed_precision_config = fsdp_config.mixed_precision
        if mixed_precision_config is not None:
            param_dtype = PrecisionType.to_dtype(mixed_precision_config.param_dtype)
            reduce_dtype = PrecisionType.to_dtype(mixed_precision_config.reduce_dtype)
            buffer_dtype = PrecisionType.to_dtype(mixed_precision_config.buffer_dtype)
        else:
            param_dtype = torch.bfloat16
            reduce_dtype = torch.float32
            buffer_dtype = torch.float32

        mixed_precision = MixedPrecision(param_dtype=param_dtype, reduce_dtype=reduce_dtype, buffer_dtype=buffer_dtype)

        auto_wrap_policy = get_fsdp_wrap_policy(module=critic_module, config=self.config.model.fsdp_config.wrap_policy, is_lora=self.config.model.lora_rank > 0)

        log_gpu_memory_usage("Before critic FSDP", logger=logger)

        fsdp_mesh = self.device_mesh
        sharding_strategy = get_sharding_strategy(fsdp_mesh)

        # Note: We force turn off CPUOffload for critic because it causes incorrect results when using grad accumulation
        if config.strategy == "fsdp":
            critic_module = FSDP(
                critic_module,
                param_init_fn=init_fn,
                use_orig_params=False,
                auto_wrap_policy=auto_wrap_policy,
                device_id=get_device_id(),
                sharding_strategy=sharding_strategy,
                mixed_precision=mixed_precision,
                sync_module_states=True,
                forward_prefetch=False,
                device_mesh=self.device_mesh,
                cpu_offload=None,
            )
        elif config.strategy == "fsdp2":
            assert CPUOffloadPolicy is not None, "PyTorch version >= 2.4 is required for using fully_shard API (FSDP2)"
            mp_policy = MixedPrecisionPolicy(param_dtype=param_dtype, reduce_dtype=reduce_dtype, cast_forward_inputs=True)
            offload_policy = None
            if fsdp_config.offload_policy:
                self._is_offload_param = False
                self._is_offload_optimizer = False
                offload_policy = CPUOffloadPolicy(pin_memory=True)

            fsdp_kwargs = {
                "mesh": fsdp_mesh,
                "mp_policy": mp_policy,
                "offload_policy": offload_policy,
                "reshard_after_forward": fsdp_config.reshard_after_forward,
            }
            full_state = critic_module.state_dict()
            apply_fsdp2(critic_module, fsdp_kwargs, fsdp_config)
            fsdp2_load_full_state_dict(critic_module, full_state, fsdp_mesh, offload_policy)
        else:
            raise NotImplementedError(f"Unknown strategy {config.strategy}")

        if config.model.enable_activation_offload:
            enable_gradient_checkpointing = config.model.enable_gradient_checkpointing
            enable_activation_offloading(critic_module, config.strategy, enable_gradient_checkpointing)

        log_gpu_memory_usage("After critic FSDP", logger=logger)

        critic_optimizer = optim.AdamW(
            critic_module.parameters(),
            lr=config.optim.lr,
            betas=config.optim.betas,
            weight_decay=config.optim.weight_decay,
        )

        total_steps = config.optim.total_training_steps
        num_warmup_steps = int(config.optim.lr_warmup_steps)
        warmup_style = config.optim.warmup_style
        if num_warmup_steps < 0:
            num_warmup_steps_ratio = config.optim.lr_warmup_steps_ratio
            num_warmup_steps = int(num_warmup_steps_ratio * total_steps)

        if self.rank == 0:
            logger.info(f"Total steps: {total_steps}, num_warmup_steps: {num_warmup_steps}")

        from siirl.utils.model_utils.torch_functional import get_constant_schedule_with_warmup, get_cosine_schedule_with_warmup

        if warmup_style == "constant":
            critic_lr_scheduler = get_constant_schedule_with_warmup(optimizer=critic_optimizer, num_warmup_steps=num_warmup_steps)
        elif warmup_style == "cosine":
            critic_lr_scheduler = get_cosine_schedule_with_warmup(optimizer=critic_optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=total_steps)
        else:
            raise NotImplementedError(f"Warmup style {warmup_style} is not supported")

        return critic_module, critic_optimizer, critic_lr_scheduler

    def init_model(self):
        # This is used to import external_lib into the huggingface systems
        import_external_libs(self.config.model.external_lib)

        from siirl.workers.critic import DataParallelPPOCritic

        self.critic_module, self.critic_optimizer, self.critic_lr_scheduler = self._build_critic_model_optimizer(self.config)

        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.critic_module)
            log_gpu_memory_usage("After offload critic model during init", logger=logger)
        if self._is_offload_optimizer:
            offload_fsdp_optimizer(optimizer=self.critic_optimizer)
            log_gpu_memory_usage("After offload critic optimizer during init", logger=logger)

        self.critic = DataParallelPPOCritic(config=self.config, critic_module=self.critic_module, critic_optimizer=self.critic_optimizer)

        self.flops_counter = FlopsCounter(self.critic_model_config)
        self.checkpoint_manager = FSDPCheckpointManager(
            model=self.critic_module, optimizer=self.critic_optimizer, lr_scheduler=self.critic_lr_scheduler, processing_class=self.processor if self.processor is not None else self.tokenizer, checkpoint_contents=self.config.checkpoint.contents, tokenizer=self.tokenizer
        )

    def compute_values(self, data: DataProto):
        # Support all hardwares
        data = data.to(get_device_id())

        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.critic_module)
        micro_batch_size = self.config.forward_micro_batch_size_per_gpu
        data.meta_info["micro_batch_size"] = micro_batch_size
        data.meta_info["max_token_len"] = self.config.forward_max_token_len_per_gpu
        data.meta_info["use_dynamic_bsz"] = self.config.use_dynamic_bsz
        # perform forward computation
        with self.ulysses_sharding_manager:
            data = self.ulysses_sharding_manager.preprocess_data(data=data)
            values = self.critic.compute_values(data=data)
            data.batch["values"] = values
            processed_data = self.ulysses_sharding_manager.postprocess_data(data=data)

        processed_data = processed_data.to("cpu")
        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.critic_module)
        return processed_data

    def update_critic(self, data: DataProto):
        # Support all hardwares
        data = data.to(get_device_id())
        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.critic_module)
        if self._is_offload_optimizer:
            load_fsdp_optimizer(optimizer=self.critic_optimizer, device_id=get_device_id())

        # perform forward computation
        with self.ulysses_sharding_manager:
            data = self.ulysses_sharding_manager.preprocess_data(data=data)

            with Timer(name="update_critic", logger=None) as timer:
                metrics = self.critic.update_critic(data=data)
            delta_time = timer.last

            global_num_tokens = data.meta_info["global_token_num"]
            estimated_flops, promised_flops = self.flops_counter.estimate_flops(global_num_tokens, delta_time)
            metrics["perf/mfu/critic"] = estimated_flops * self.config.ppo_epochs / promised_flops

            self.critic_lr_scheduler.step()
            lr = self.critic_lr_scheduler.get_last_lr()[0]
            metrics["critic/lr"] = lr

            data.meta_info["metrics"] = metrics
            processed_data = self.ulysses_sharding_manager.postprocess_data(data=data)

        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.critic_module)
        if self._is_offload_optimizer:
            offload_fsdp_optimizer(optimizer=self.critic_optimizer)

        processed_data = processed_data.to("cpu")
        return processed_data

    def save_checkpoint(self, local_path, hdfs_path=None, global_step=0, max_ckpt_to_keep=None):
        import torch

        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.critic_module)

        self.checkpoint_manager.save_checkpoint(local_path=local_path, hdfs_path=hdfs_path, global_step=global_step, max_ckpt_to_keep=max_ckpt_to_keep)

        torch.distributed.barrier()
        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.critic_module)

    def load_checkpoint(self, local_path, hdfs_path=None, del_local_after_load=True):
        import torch

        if self._is_offload_param:
            load_fsdp_model_to_gpu(self.critic_module)

        self.checkpoint_manager.load_checkpoint(local_path=local_path, hdfs_path=hdfs_path, del_local_after_load=del_local_after_load)

        torch.distributed.barrier()
        if self._is_offload_param:
            offload_fsdp_model_to_cpu(self.critic_module)

        if self._is_offload_optimizer:
            offload_fsdp_optimizer(self.critic_optimizer)


# TODO(sgm): we may need to extract it to dp_reward_model.py
class RewardModelWorker(Worker):
    """
    Note that we only implement the reward model that is subclass of AutoModelForTokenClassification.
    """

    def __init__(self, config: RewardModelArguments, process_group: ProcessGroup):
        super().__init__()
        import torch.distributed

        if not torch.distributed.is_initialized():
            torch.distributed.init_process_group(backend=get_nccl_backend())
        self.config = config

        # build device mesh for Ulysses Sequence Parallel
        world_size = torch.distributed.get_world_size(group=process_group)

        self.device_mesh = create_device_mesh_from_group(process_group=process_group, fsdp_size=self.config.model.fsdp_config.fsdp_size)

        self.ulysses_device_mesh = None
        self.ulysses_sequence_parallel_size = self.config.ulysses_sequence_parallel_size
        if self.ulysses_sequence_parallel_size > 1:
            self.ulysses_device_mesh = create_device_mesh_from_group(process_group=process_group, sp_size=self.ulysses_sequence_parallel_size)

        self.ulysses_sharding_manager = FSDPUlyssesShardingManager(self.ulysses_device_mesh)

        self.use_remove_padding = self.config.model.use_remove_padding

        # normalize config
        if self.config.micro_batch_size is not None:
            self.config.micro_batch_size //= world_size
            self.config.micro_batch_size_per_gpu = self.config.micro_batch_size

    def _build_model(self, config: RewardModelArguments):
        # the following line is necessary
        from torch.distributed.fsdp import CPUOffload
        from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
        from transformers import AutoConfig, AutoModelForTokenClassification

        use_shm = config.model.use_shm
        # download the checkpoint from hdfs
        local_path = copy_to_local(config.model.path, use_shm=use_shm)

        if self.config.model.input_tokenizer is None:
            self._do_switch_chat_template = False
        else:
            self._do_switch_chat_template = True
            input_tokenizer_local_path = copy_to_local(config.model.input_tokenizer)
            input_tokenizer_module = load_tokenizer(path=input_tokenizer_local_path)
            self.input_tokenizer = input_tokenizer_module["tokenizer"]
            tokenizer_module = load_tokenizer(model_args=self.config.model)
            self.tokenizer, self.processor = tokenizer_module["tokenizer"], tokenizer_module["processor"]

        trust_remote_code = config.model.trust_remote_code
        model_config = AutoConfig.from_pretrained(local_path, trust_remote_code=trust_remote_code)
        model_config.num_labels = 1

        # note that we have to create model in fp32. Otherwise, the optimizer is in bf16, which is incorrect
        init_context = get_init_weight_context_manager(use_meta_tensor=not model_config.tie_word_embeddings, mesh=self.device_mesh)

        with init_context(), warnings.catch_warnings():
            warnings.simplefilter("ignore")
            model_config.classifier_dropout = 0.0
            reward_module = AutoModelForTokenClassification.from_pretrained(
                pretrained_model_name_or_path=local_path,
                config=model_config,
                torch_dtype=torch.bfloat16,
                attn_implementation="flash_attention_2",
                trust_remote_code=trust_remote_code,
            )

            apply_monkey_patch(
                model=reward_module,
                use_remove_padding=config.model.use_remove_padding,
                ulysses_sp_size=self.ulysses_sequence_parallel_size,
            )

            reward_module.to(torch.bfloat16)

        auto_wrap_policy = get_fsdp_wrap_policy(module=reward_module, config=self.config.model.fsdp_config.wrap_policy)

        fsdp_mesh = self.device_mesh
        sharding_strategy = get_sharding_strategy(fsdp_mesh)

        if config.strategy == "fsdp":
            reward_module = FSDP(
                reward_module,
                param_init_fn=init_fn,
                use_orig_params=False,
                auto_wrap_policy=auto_wrap_policy,
                device_id=get_device_id(),
                sharding_strategy=sharding_strategy,  # zero3
                sync_module_states=True,
                cpu_offload=CPUOffload(offload_params=True),
                forward_prefetch=False,
                device_mesh=self.device_mesh,
            )
        elif config.strategy == "fsdp2":
            assert CPUOffloadPolicy is not None, "PyTorch version >= 2.4 is required for using fully_shard API (FSDP2)"
            cpu_offload = CPUOffloadPolicy(pin_memory=True)
            fsdp_kwargs = {
                "mesh": fsdp_mesh,
                "offload_policy": cpu_offload,
                "reshard_after_forward": config.model.fsdp_config.reshard_after_forward,
            }
            full_state = reward_module.state_dict()
            apply_fsdp2(reward_module, fsdp_kwargs, config.model.fsdp_config)
            fsdp2_load_full_state_dict(reward_module, full_state, fsdp_mesh, cpu_offload)
        else:
            raise NotImplementedError(f"Unknown strategy: {config.strategy}")
        return reward_module

    def init_model(self):
        # This is used to import external_lib into the huggingface systems
        import_external_libs(self.config.model.external_lib)
        self.reward_module = self._build_model(config=self.config)

    def _forward_micro_batch(self, micro_batch):
        if is_cuda_available:
            from flash_attn.bert_padding import index_first_axis, pad_input, rearrange, unpad_input
        elif is_npu_available:
            from transformers.integrations.npu_flash_attention import index_first_axis, pad_input, rearrange, unpad_input

        from siirl.utils.model_utils.ulysses import gather_outpus_and_unpad, ulysses_pad_and_slice_inputs

        with torch.no_grad(), torch.autocast(device_type=device_name, dtype=torch.bfloat16):
            input_ids = micro_batch["input_ids"]
            batch_size, seqlen = input_ids.shape
            attention_mask = micro_batch["attention_mask"]
            position_ids = micro_batch["position_ids"]

            if self.use_remove_padding:
                input_ids_rmpad, indices, *_ = unpad_input(input_ids.unsqueeze(-1), attention_mask)  # input_ids_rmpad (total_nnz, ...)
                input_ids_rmpad = input_ids_rmpad.transpose(0, 1)  # (1, total_nnz)

                # unpad the position_ids to align the rotary
                position_ids_rmpad = index_first_axis(rearrange(position_ids.unsqueeze(-1), "b s ... -> (b s) ..."), indices).transpose(0, 1)

                # pad and slice the inputs if sp > 1
                if self.ulysses_sequence_parallel_size > 1:
                    input_ids_rmpad, position_ids_rmpad, pad_size = ulysses_pad_and_slice_inputs(input_ids_rmpad, position_ids_rmpad, sp_size=self.ulysses_sequence_parallel_size)

                # only pass input_ids and position_ids to enable flash_attn_varlen
                output = self.reward_module(input_ids=input_ids_rmpad, attention_mask=None, position_ids=position_ids_rmpad, use_cache=False)  # prevent model thinks we are generating
                reward_rmpad = output.logits
                reward_rmpad = reward_rmpad.squeeze(0)  # (total_nnz)

                # gather output if sp > 1
                if self.ulysses_sequence_parallel_size > 1:
                    reward_rmpad = gather_outpus_and_unpad(reward_rmpad, gather_dim=0, unpad_dim=0, padding_size=pad_size)

                # pad it back
                rm_score = pad_input(reward_rmpad, indices=indices, batch=batch_size, seqlen=seqlen).squeeze(-1)
            else:
                output = self.reward_module(input_ids=input_ids, attention_mask=attention_mask, position_ids=position_ids, use_cache=False)
                rm_score = output.logits  # (batch_size, seq_len, 1)
                rm_score = rm_score.squeeze(-1)

            # extract the result of the last valid token
            eos_mask_idx = torch.argmax(position_ids * attention_mask, dim=-1)  # (bsz,)
            rm_score = rm_score[torch.arange(batch_size), eos_mask_idx]
            return rm_score

    def _expand_to_token_level(self, data: DataProto, scores: torch.Tensor):
        batch_size = data.batch.batch_size[0]
        # expand as token_level_reward
        attention_mask = data.batch["attention_mask"]
        position_ids = data.batch["position_ids"]
        response_length = data.batch["responses"].shape[-1]
        eos_mask_idx = torch.argmax(position_ids * attention_mask, dim=-1)  # (bsz,)
        token_level_scores = torch.zeros_like(attention_mask, dtype=scores.dtype)  # (bsz, seqlen)
        token_level_scores[torch.arange(batch_size), eos_mask_idx] = scores

        # select the response part
        token_level_scores = token_level_scores[:, -response_length:]

        return token_level_scores

    def _switch_chat_template(self, data: DataProto):
        src_max_length = data.batch["attention_mask"].shape[-1]

        src_tokenizer = self.input_tokenizer
        target_tokenizer = self.tokenizer

        rm_input_ids = []
        rm_attention_mask = []

        for i in range(data.batch.batch_size[0]):
            # extract raw prompt
            if isinstance(data.non_tensor_batch["raw_prompt"][i], list):
                chat: list = data.non_tensor_batch["raw_prompt"][i]
            else:
                chat: list = data.non_tensor_batch["raw_prompt"][i].tolist()

            # extract response
            response_ids = data.batch["responses"][i]
            response_length = response_ids.shape[-1]
            valid_response_length = data.batch["attention_mask"][i][-response_length:].sum()
            valid_response_ids = response_ids[:valid_response_length]

            # decode
            response = src_tokenizer.decode(valid_response_ids)
            # remove bos and eos
            response = response.replace(src_tokenizer.eos_token, "")

            chat.append({"role": "assistant", "content": response})

            prompt_with_chat_template = target_tokenizer.apply_chat_template(chat, add_generation_prompt=False, tokenize=False)
            if self.rank == 0 and i == 0:
                # for debugging purpose
                logger.info(f"Switch template. chat: {prompt_with_chat_template}")

            # the maximum length is actually determined by the reward model itself
            max_length = self.config.max_length
            if max_length is None:
                max_length = src_max_length

            model_inputs = target_tokenizer(prompt_with_chat_template, return_tensors="pt", add_special_tokens=False)
            input_ids, attention_mask = F.postprocess_data(
                input_ids=model_inputs["input_ids"],
                attention_mask=model_inputs["attention_mask"],
                max_length=max_length,
                pad_token_id=target_tokenizer.pad_token_id,
                left_pad=False,  # right padding
                truncation=self.config.truncation,
            )  # truncate from the right

            rm_input_ids.append(input_ids)
            rm_attention_mask.append(attention_mask)

        rm_input_ids = torch.cat(rm_input_ids, dim=0)
        rm_attention_mask = torch.cat(rm_attention_mask, dim=0)

        rm_position_ids = compute_position_id_with_mask(rm_attention_mask)

        rm_inputs = {"input_ids": rm_input_ids, "attention_mask": rm_attention_mask, "position_ids": rm_position_ids}

        return DataProto.from_dict(rm_inputs)

    def compute_rm_score(self, data: DataProto):
        import itertools

        from siirl.utils.model_utils.seqlen_balancing import get_reverse_idx, rearrange_micro_batches

        # Support all hardwares
        data = data.to(get_device_id())
        if self._do_switch_chat_template:
            rm_data = self._switch_chat_template(data)
        else:
            rm_input_ids = data.batch["input_ids"]
            rm_attention_mask = data.batch["attention_mask"]
            rm_position_ids = data.batch["position_ids"]
            rm_inputs = {
                "input_ids": rm_input_ids,
                "attention_mask": rm_attention_mask,
                "position_ids": rm_position_ids,
            }
            rm_data = DataProto.from_dict(rm_inputs)

        # Support all hardwares
        rm_data.batch = rm_data.batch.to(get_device_id())

        # perform forward computation
        with self.ulysses_sharding_manager:
            rm_data = self.ulysses_sharding_manager.preprocess_data(data=rm_data)
            data = self.ulysses_sharding_manager.preprocess_data(data=data)

            use_dynamic_bsz = self.config.use_dynamic_bsz
            if use_dynamic_bsz:
                max_token_len = self.config.forward_max_token_len_per_gpu * self.ulysses_sequence_parallel_size
                micro_batches, indices = rearrange_micro_batches(batch=rm_data.batch, max_token_len=max_token_len)
            else:
                micro_batches = rm_data.batch.split(self.config.micro_batch_size_per_gpu)
            output = []
            for micro_batch in micro_batches:
                rm_score = self._forward_micro_batch(micro_batch)
                output.append(rm_score)
            scores = torch.cat(output, dim=0)  # (batch_size)

            if use_dynamic_bsz:
                indices = list(itertools.chain.from_iterable(indices))
                assert len(indices) == scores.size(0), f"{len(indices)} vs. {scores.size()}"
                revert_indices = torch.tensor(get_reverse_idx(indices), dtype=torch.long)
                scores = scores[revert_indices]

            token_level_scores = self._expand_to_token_level(data, scores)
            # Note that this is only the scores, may not be the final rewards used to train RL
            data.batch["rm_scores"] = token_level_scores
            processed_data = self.ulysses_sharding_manager.postprocess_data(data=data)

        # https://pytorch.org/docs/stable/notes/fsdp.html#fsdp-notes
        # unshard the root FSDP module
        if self.world_size > 1 and fsdp_version(self.reward_module) == 1:
            self.reward_module._handle.reshard(True)

        processed_data = processed_data.to("cpu")
        return processed_data


# ================================= Async related workers =================================
class AsyncActorRolloutRefWorker(ActorRolloutRefWorker):
    def _build_rollout(self, trust_remote_code=False):
        rollout, rollout_sharding_manager = super()._build_rollout(trust_remote_code)

        # NOTE: rollout is not actually initialized here, it's deferred
        # to be initialized by AsyncvLLMServer.

        self.vllm_tp_size = self.config.rollout.tensor_model_parallel_size
        self.vllm_dp_rank = int(os.environ["RANK"]) // self.vllm_tp_size
        self.vllm_tp_rank = int(os.environ["RANK"]) % self.vllm_tp_size

        # used for sleep/wake_up
        rollout.sharding_manager = rollout_sharding_manager

        return rollout, rollout_sharding_manager

    def generate_sequences(self, prompts: DataProto):
        raise NotImplementedError("AsyncActorRolloutRefWorker does not support generate_sequences")

    def execute_method(self, method: Union[str, bytes], *args, **kwargs):
        """Called by ExternalRayDistributedExecutor collective_rpc."""
        if self.vllm_tp_rank == 0 and method != "execute_model":
            print(f"[DP={self.vllm_dp_rank},TP={self.vllm_tp_rank}] execute_method: {method if isinstance(method, str) else 'Callable'}")
        return self.rollout.execute_method(method, *args, **kwargs)

    async def chat_completion(self, json_request):
        ret = await self.rollout.chat_completion(json_request)
        return ret

    async def wake_up(self):
        await self.rollout.wake_up()
        # return something to block the caller
        return True

    async def sleep(self):
        await self.rollout.sleep()
        # return something to block the caller
        return True
    
    def set_rollout_sharding_manager(self, sharding_manager):
        super().set_rollout_sharding_manager(sharding_manager)
        self.rollout.sharding_manager = sharding_manager
                
    def get_zeromq_address(self):
        return self.rollout.get_zeromq_address()




# --------------------------------------------------------
# InternVL
# Copyright (c) 2024 OpenGVLab
# Licensed under The MIT License [see LICENSE for details]
# --------------------------------------------------------

import warnings
from typing import List, Optional, Tuple, Union

import torch.utils.checkpoint
import transformers
from torch import nn
from torch.nn import CrossEntropyLoss
from transformers import AutoModel, GenerationConfig, LlamaForCausalLM, LlamaTokenizer
from transformers.modeling_outputs import CausalLMOutputWithPast
from transformers.modeling_utils import PreTrainedModel
from transformers.utils import ModelOutput, logging

from .configuration_internvl_chat import InternVLChatConfig
from .modeling_intern_vit import InternVisionModel, has_flash_attn
from .modeling_internlm2 import InternLM2ForCausalLM

from siirl.models.transformers.internvl import get_conv_template

logger = logging.get_logger(__name__)


def version_cmp(v1, v2, op="eq"):
    import operator

    from packaging import version

    op_func = getattr(operator, op)
    return op_func(version.parse(v1), version.parse(v2))


class InternVLChatModel(PreTrainedModel):
    config_class = InternVLChatConfig
    main_input_name = "pixel_values"
    base_model_prefix = "language_model"
    _supports_flash_attn_2 = True
    _no_split_modules = ["InternVisionEncoderLayer", "LlamaDecoderLayer", "InternLM2DecoderLayer"]
    supports_gradient_checkpointing = True

    def __init__(self, config: InternVLChatConfig, vision_model=None, language_model=None, use_flash_attn=True):
        super().__init__(config)

        assert version_cmp(transformers.__version__, "4.37.0", "ge")
        image_size = config.force_image_size or config.vision_config.image_size
        patch_size = config.vision_config.patch_size
        self.patch_size = patch_size
        self.select_layer = config.select_layer
        self.template = config.template
        self.num_image_token = int((image_size // patch_size) ** 2 * (config.downsample_ratio**2))
        self.downsample_ratio = config.downsample_ratio
        self.ps_version = config.ps_version
        use_flash_attn = use_flash_attn if has_flash_attn else False
        config.vision_config.use_flash_attn = True if use_flash_attn else False
        config.llm_config.attn_implementation = "flash_attention_2" if use_flash_attn else "eager"

        logger.info(f"num_image_token: {self.num_image_token}")
        logger.info(f"ps_version: {self.ps_version}")
        if vision_model is not None:
            self.vision_model = vision_model
        else:
            self.vision_model = InternVisionModel(config.vision_config)
        if language_model is not None:
            self.language_model = language_model
        else:
            if config.llm_config.architectures[0] == "LlamaForCausalLM":
                self.language_model = LlamaForCausalLM(config.llm_config)
            elif config.llm_config.architectures[0] == "InternLM2ForCausalLM":
                self.language_model = InternLM2ForCausalLM(config.llm_config)
            else:
                raise NotImplementedError(f"{config.llm_config.architectures[0]} is not implemented.")

        vit_hidden_size = config.vision_config.hidden_size
        llm_hidden_size = config.llm_config.hidden_size

        self.mlp1 = nn.Sequential(nn.LayerNorm(vit_hidden_size * int(1 / self.downsample_ratio) ** 2), nn.Linear(vit_hidden_size * int(1 / self.downsample_ratio) ** 2, llm_hidden_size), nn.GELU(), nn.Linear(llm_hidden_size, llm_hidden_size))

        self.img_context_token_id = None
        self.conv_template = get_conv_template(self.template)
        self.system_message = self.conv_template.system_message

    def forward(
        self,
        pixel_values: torch.FloatTensor,
        input_ids: torch.LongTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        image_flags: Optional[torch.LongTensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
        verbose: bool = False,
    ) -> Union[Tuple, CausalLMOutputWithPast]:
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        image_flags = image_flags.squeeze(-1)
        input_embeds = self.language_model.get_input_embeddings()(input_ids).clone()

        vit_embeds = self.extract_feature(pixel_values)
        vit_embeds = vit_embeds[image_flags == 1]
        vit_batch_size = pixel_values.shape[0]

        B, N, C = input_embeds.shape
        input_embeds = input_embeds.reshape(B * N, C)

        if verbose and torch.distributed.is_initialized() and torch.distributed.get_rank() == 0:
            print(f"dynamic ViT batch size: {vit_batch_size}, images per sample: {vit_batch_size / B}, dynamic token length: {N}")

        input_ids = input_ids.reshape(B * N)
        selected = input_ids == self.img_context_token_id
        try:
            input_embeds[selected] = input_embeds[selected] * 0.0 + vit_embeds.reshape(-1, C)
        except Exception as e:
            vit_embeds = vit_embeds.reshape(-1, C)
            print(f"warning: {e}, input_embeds[selected].shape={input_embeds[selected].shape}, vit_embeds.shape={vit_embeds.shape}")
            n_token = selected.sum()
            pad_size = n_token - vit_embeds.size(0)
            if pad_size > 0:
                vit_embeds = nn.functional.pad(vit_embeds, (0, 0, 0, pad_size))
            input_embeds[selected] = input_embeds[selected] * 0.0 + vit_embeds[:n_token]

        input_embeds = input_embeds.reshape(B, N, C)

        outputs = self.language_model(
            inputs_embeds=input_embeds,
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )
        logits = outputs.logits

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = logits[..., :-1, :].contiguous()
            shift_labels = labels[..., 1:].contiguous()
            # Flatten the tokens
            loss_fct = CrossEntropyLoss()
            shift_logits = shift_logits.view(-1, self.language_model.config.vocab_size)
            shift_labels = shift_labels.view(-1)
            # Enable model parallelism
            shift_labels = shift_labels.to(shift_logits.device)
            loss = loss_fct(shift_logits, shift_labels)

        if not return_dict:
            output = (logits,) + outputs[1:]
            return (loss,) + output if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def pixel_shuffle(self, x, scale_factor=0.5):
        n, w, h, c = x.size()
        # N, W, H, C --> N, W, H * scale, C // scale
        x = x.view(n, w, int(h * scale_factor), int(c / scale_factor))
        # N, W, H * scale, C // scale --> N, H * scale, W, C // scale
        x = x.permute(0, 2, 1, 3).contiguous()
        # N, H * scale, W, C // scale --> N, H * scale, W * scale, C // (scale ** 2)
        x = x.view(n, int(h * scale_factor), int(w * scale_factor), int(c / (scale_factor * scale_factor)))
        if self.ps_version == "v1":
            warnings.warn("In ps_version 'v1', the height and width have not been swapped back, which results in a transposed image.")
        else:
            x = x.permute(0, 2, 1, 3).contiguous()
        return x

    def extract_feature(self, pixel_values):
        if self.select_layer == -1:
            vit_embeds = self.vision_model(pixel_values=pixel_values, output_hidden_states=False, return_dict=True).last_hidden_state
        else:
            vit_embeds = self.vision_model(pixel_values=pixel_values, output_hidden_states=True, return_dict=True).hidden_states[self.select_layer]
        vit_embeds = vit_embeds[:, 1:, :]

        h = w = int(vit_embeds.shape[1] ** 0.5)
        vit_embeds = vit_embeds.reshape(vit_embeds.shape[0], h, w, -1)
        vit_embeds = self.pixel_shuffle(vit_embeds, scale_factor=self.downsample_ratio)
        vit_embeds = vit_embeds.reshape(vit_embeds.shape[0], -1, vit_embeds.shape[-1])
        vit_embeds = self.mlp1(vit_embeds)
        return vit_embeds

    def batch_chat(self, tokenizer, pixel_values, questions, generation_config, num_patches_list=None, history=None, return_history=False, IMG_START_TOKEN="<img>", IMG_END_TOKEN="</img>", IMG_CONTEXT_TOKEN="<IMG_CONTEXT>", verbose=False, image_counts=None):
        if history is not None or return_history:
            print("Now multi-turn chat is not supported in batch_chat.")
            raise NotImplementedError

        if image_counts is not None:
            num_patches_list = image_counts
            print("Warning: `image_counts` is deprecated. Please use `num_patches_list` instead.")

        img_context_token_id = tokenizer.convert_tokens_to_ids(IMG_CONTEXT_TOKEN)
        self.img_context_token_id = img_context_token_id

        if verbose and pixel_values is not None:
            image_bs = pixel_values.shape[0]
            print(f"dynamic ViT batch size: {image_bs}")

        queries = []
        for idx, num_patches in enumerate(num_patches_list):
            question = questions[idx]
            if pixel_values is not None and "<image>" not in question:
                question = "<image>\n" + question
            template = get_conv_template(self.template)
            template.system_message = self.system_message
            template.append_message(template.roles[0], question)
            template.append_message(template.roles[1], None)
            query = template.get_prompt()

            image_tokens = IMG_START_TOKEN + IMG_CONTEXT_TOKEN * self.num_image_token * num_patches + IMG_END_TOKEN
            query = query.replace("<image>", image_tokens, 1)
            queries.append(query)

        tokenizer.padding_side = "left"
        model_inputs = tokenizer(queries, return_tensors="pt", padding=True)
        input_ids = model_inputs["input_ids"].to(self.device)
        attention_mask = model_inputs["attention_mask"].to(self.device)
        eos_token_id = tokenizer.convert_tokens_to_ids(template.sep.strip())
        generation_config["eos_token_id"] = eos_token_id
        generation_output = self.generate(pixel_values=pixel_values, input_ids=input_ids, attention_mask=attention_mask, **generation_config)
        responses = tokenizer.batch_decode(generation_output, skip_special_tokens=True)
        responses = [response.split(template.sep.strip())[0].strip() for response in responses]
        return responses

    def chat(self, tokenizer, pixel_values, question, generation_config, history=None, return_history=False, num_patches_list=None, IMG_START_TOKEN="<img>", IMG_END_TOKEN="</img>", IMG_CONTEXT_TOKEN="<IMG_CONTEXT>", verbose=False):
        if history is None and pixel_values is not None and "<image>" not in question:
            question = "<image>\n" + question

        if num_patches_list is None:
            num_patches_list = [pixel_values.shape[0]] if pixel_values is not None else []
        assert pixel_values is None or len(pixel_values) == sum(num_patches_list)

        img_context_token_id = tokenizer.convert_tokens_to_ids(IMG_CONTEXT_TOKEN)
        self.img_context_token_id = img_context_token_id

        template = get_conv_template(self.template)
        template.system_message = self.system_message
        eos_token_id = tokenizer.convert_tokens_to_ids(template.sep.strip())

        history = [] if history is None else history
        for old_question, old_answer in history:
            template.append_message(template.roles[0], old_question)
            template.append_message(template.roles[1], old_answer)
        template.append_message(template.roles[0], question)
        template.append_message(template.roles[1], None)
        query = template.get_prompt()

        if verbose and pixel_values is not None:
            image_bs = pixel_values.shape[0]
            print(f"dynamic ViT batch size: {image_bs}")

        for num_patches in num_patches_list:
            image_tokens = IMG_START_TOKEN + IMG_CONTEXT_TOKEN * self.num_image_token * num_patches + IMG_END_TOKEN
            query = query.replace("<image>", image_tokens, 1)

        model_inputs = tokenizer(query, return_tensors="pt")
        input_ids = model_inputs["input_ids"].to(self.device)
        attention_mask = model_inputs["attention_mask"].to(self.device)
        generation_config["eos_token_id"] = eos_token_id
        generation_output = self.generate(pixel_values=pixel_values, input_ids=input_ids, attention_mask=attention_mask, **generation_config)
        response = tokenizer.batch_decode(generation_output, skip_special_tokens=True)[0]
        response = response.split(template.sep.strip())[0].strip()
        history.append((question, response))
        if return_history:
            return response, history
        else:
            query_to_print = query.replace(IMG_CONTEXT_TOKEN, "")
            query_to_print = query_to_print.replace(f"{IMG_START_TOKEN}{IMG_END_TOKEN}", "<image>")
            if verbose:
                print(query_to_print, response)
            return response

    @torch.no_grad()
    def generate(
        self,
        pixel_values: Optional[torch.FloatTensor] = None,
        input_ids: Optional[torch.FloatTensor] = None,
        attention_mask: Optional[torch.LongTensor] = None,
        visual_features: Optional[torch.FloatTensor] = None,
        generation_config: Optional[GenerationConfig] = None,
        output_hidden_states: Optional[bool] = None,
        **generate_kwargs,
    ) -> torch.LongTensor:
        assert self.img_context_token_id is not None
        if pixel_values is not None:
            if visual_features is not None:
                vit_embeds = visual_features
            else:
                vit_embeds = self.extract_feature(pixel_values)
            input_embeds = self.language_model.get_input_embeddings()(input_ids)
            B, N, C = input_embeds.shape
            input_embeds = input_embeds.reshape(B * N, C)

            input_ids = input_ids.reshape(B * N)
            selected = input_ids == self.img_context_token_id
            assert selected.sum() != 0
            input_embeds[selected] = vit_embeds.reshape(-1, C).to(input_embeds.device)

            input_embeds = input_embeds.reshape(B, N, C)
        else:
            input_embeds = self.language_model.get_input_embeddings()(input_ids)

        outputs = self.language_model.generate(
            inputs_embeds=input_embeds,
            attention_mask=attention_mask,
            generation_config=generation_config,
            output_hidden_states=output_hidden_states,
            use_cache=True,
            **generate_kwargs,
        )

        return outputs



# Copyright 2023-2024 SGLang Team
# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import asyncio
import logging
from typing import Any, Dict, List, Tuple
import pickle
import zmq
import torch
import ray
from omegaconf import DictConfig
from starlette.requests import Request
from starlette.responses import JSONResponse

from siirl.workers.rollout.async_server import AsyncServerBase
from siirl.utils.params.model_args import ActorRolloutRefArguments
logger = logging.getLogger(__file__)


@ray.remote(num_cpus=1)
class AsyncSglangServer(AsyncServerBase):
    def __init__(self, config: ActorRolloutRefArguments, global_rank: int, wg_prefix: str):
        super().__init__()
        self.config = config.rollout
        self.global_rank = global_rank
        self.wg_prefix = wg_prefix
        self.workers_zmq = []
        self.master_worker_zmq = None
        
    async def init_engine(self):
        if self.workers_zmq:
            # avoid init twice
            return
        actor_names = [
            actor_name for actor_name in ray.util.list_named_actors() if actor_name.startswith(f"{self.wg_prefix}_DAGWorker")
        ]
        vllm_tp_size = self.config.tensor_model_parallel_size
        def get_pg_index_and_local_rank(actor_name) -> Tuple[int, int]:
            fields = actor_name.split(":")
            assert len(fields) == 2, f"invalid actor name: {actor_name}"
            pg_index, local_rank = int(fields[0].split("_")[-1]), int(fields[1])
            return pg_index, local_rank
    # sort actor names by pg_index and local_rank
        actor_names = sorted(actor_names, key=get_pg_index_and_local_rank)
        actor_names = actor_names[self.global_rank : self.global_rank + vllm_tp_size]
        matched_actors = [
            ray.get_actor(actor_name) for actor_name in actor_names
        ]
        self.context = zmq.Context()
        for worker in matched_actors:
            zmq_address = ray.get(worker.get_zeromq_address.remote())
            socket = self.context.socket(zmq.REQ)
            socket.connect(zmq_address)
            self.workers_zmq.append(socket)
        self.master_worker_zmq = self.workers_zmq[0]

    async def chat_completion(self, raw_request: Request):
        request = await raw_request.json()
        message = pickle.dumps(('chat_completion', (), {'request':request}))
        self.master_worker_zmq.send(message, zmq.DONTWAIT)
        outputs = []
        outputs.append(pickle.loads(self.master_worker_zmq.recv()))
        return JSONResponse(outputs)


    async def generate(self, prompt_ids: List[int], sampling_params: Dict[str, Any], request_id: str) -> List[int]:
        message = pickle.dumps(('generate', (), {'prompt_ids':prompt_ids, 'sampling_params':sampling_params, 'request_id':request_id}))
        self.master_worker_zmq.send(message, zmq.DONTWAIT)
        return pickle.loads(self.master_worker_zmq.recv())
        return await self.master_worker.generate.remote(prompt_ids, sampling_params, request_id)

    async def wake_up(self):
        if not self.config.free_cache_engine:
            return
        message = pickle.dumps(('wake_up', (), {}))
        for socket in self.workers_zmq:
            socket.send(message, zmq.DONTWAIT)
        for socket in self.workers_zmq:
            socket.recv()
        return

    async def sleep(self):
        if not self.config.free_cache_engine:
            return
        message = pickle.dumps(('sleep', (), {}))
        for socket in self.workers_zmq:
            socket.send(message, zmq.DONTWAIT)
        for socket in self.workers_zmq:
            socket.recv()
        return



# Copyright 2025 Bytedance Ltd. and/or its affiliates
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import threading
import time

from mcp import Tool

logger = logging.getLogger(__file__)


class TokenBucket:
    def __init__(self, rate_limit: float):
        self.rate_limit = rate_limit  # tokens per second
        self.tokens = rate_limit
        self.last_update = time.time()
        self.lock = threading.Lock()

    def acquire(self) -> bool:
        with self.lock:
            now = time.time()
            # Add new tokens based on time elapsed
            new_tokens = (now - self.last_update) * self.rate_limit
            self.tokens = min(self.rate_limit, self.tokens + new_tokens)
            self.last_update = now

            if self.tokens >= 1:
                self.tokens -= 1
                return True
            return False


def mcp2openai(mcp_tool: Tool) -> dict:
    """Convert a MCP Tool to an OpenAI ChatCompletionTool."""
    openai_format = {
        "type": "function",
        "function": {
            "name": mcp_tool.name,
            "description": mcp_tool.description,
            "parameters": mcp_tool.inputSchema,
            "strict": False,
        },
    }
    if not openai_format["function"]["parameters"].get("required", None):
        openai_format["function"]["parameters"]["required"] = []
    return openai_format



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Tuple, Type, Any, Optional
from dataclasses import asdict, is_dataclass

from dacite import Config as DaciteConfig, from_dict
from loguru import logger
from omegaconf import DictConfig, OmegaConf

from siirl.workers.dag import NodeRole, NodeType, TaskGraph
from siirl.utils.params import ActorRolloutRefArguments, CriticArguments, RewardModelArguments, SiiRLArguments

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from siirl.workers.dag import TaskGraph
from siirl.utils.params import log_dict_formatted

NODE_ID = "_node_id_"
INTERN_CONFIG = "intern_config"


def unflatten_dict_with_omegaconf(flat_dict: Dict[str, Any]) -> Dict[str, Any]:
    """
    Unflattens a flat dictionary with dot-separated keys into a nested dictionary using OmegaConf.

    Args:
        flat_dict: A dictionary where keys might be dot-separated (e.g., 'model.name').

    Returns:
        A nested dictionary.
    """
    if not flat_dict:
        return {}
    config = OmegaConf.create()
    for key, value in flat_dict.items():
        try:
            OmegaConf.update(config, key, value, merge=True, force_add=True)
        except Exception as e:
            logger.error(f"OmegaConf.update failed for key='{key}', value='{value}': {e}")
            raise
    return OmegaConf.to_container(config, resolve=True, throw_on_missing=False)


def update_task_graph_node_configs(workerflow_taskgraph: TaskGraph, basic_common_config: "SiiRLArguments") -> TaskGraph:
    """
    Updates node configurations by merging global defaults with node-specific overrides,
    and stores the resulting configuration as both a dictionary and a dataclass instance.

    Args:
        workerflow_taskgraph: The TaskGraph whose nodes will be updated.
        basic_common_config: The global SiiRLArguments with default settings.

    Returns:
        The updated TaskGraph.
    """
    logger.info("Starting update of TaskGraph node configurations (using OmegaConf and Dacite)...")
    workerflow_taskgraph.build_adjacency_lists()

    node_role_config_map: Dict[NodeRole, Tuple[str, Type]] = {
        NodeRole.ACTOR: ("actor_rollout_ref", ActorRolloutRefArguments),
        NodeRole.ROLLOUT: ("actor_rollout_ref", ActorRolloutRefArguments),
        NodeRole.REFERENCE: ("actor_rollout_ref", ActorRolloutRefArguments),
        NodeRole.CRITIC: ("critic", CriticArguments),
        NodeRole.REWARD: ("reward_model", RewardModelArguments),
    }

    for node_id, node in workerflow_taskgraph.nodes.items():
        if node.node_type not in [NodeType.MODEL_INFERENCE, NodeType.MODEL_TRAIN]:
            logger.debug(f"Node '{node.node_id}' of type {node.node_type} skipped for config update.")
            continue

        original_node_config_flat = node.config or {}
        original_node_config_dict = unflatten_dict_with_omegaconf(original_node_config_flat)

        if NODE_ID in original_node_config_dict:
            del original_node_config_dict[NODE_ID]
        node_specific_omega_conf = OmegaConf.create(original_node_config_dict)

        if node.node_role in node_role_config_map:
            default_config_attr_name, target_dataclass_type = node_role_config_map[node.node_role]
            default_config_branch_instance = getattr(basic_common_config, default_config_attr_name, None)

            merged_omega_conf: Optional[DictConfig] = None

            if default_config_branch_instance is None:
                logger.warning(f"Global default config '{default_config_attr_name}' not in basic_common_config for node '{node.node_id}'. Using only node-specific config.")
                merged_omega_conf = node_specific_omega_conf
            else:
                default_config_branch_dict = asdict(default_config_branch_instance)
                default_config_branch_omega_base = OmegaConf.create(default_config_branch_dict)

                if not isinstance(default_config_branch_omega_base, DictConfig):
                    logger.error(f"Global config for '{default_config_attr_name}' is not a DictConfig. Cannot merge. Using only node-specific config for node '{node.node_id}'.")
                    merged_omega_conf = node_specific_omega_conf
                else:
                    merged_omega_conf = OmegaConf.merge(default_config_branch_omega_base.copy(), node_specific_omega_conf)

            merged_config_dict = OmegaConf.to_container(merged_omega_conf, resolve=True, throw_on_missing=False)
            if not isinstance(merged_config_dict, dict):
                raise ValueError(f"Merged config for node '{node.node_id}' is not a dictionary.")

            try:
                # Convert the merged dictionary back into a validated dataclass instance
                merged_dataclass_instance = from_dict(data_class=target_dataclass_type, data=merged_config_dict, config=DaciteConfig(check_types=False))
            except Exception as e:
                logger.error(f"Dacite conversion to '{target_dataclass_type.__name__}' failed for node '{node.node_id}': {e}")
                raise
            node.config = {INTERN_CONFIG: merged_dataclass_instance, NODE_ID: node.node_id}

        else:
            logger.warning(f"Node '{node.node_id}' ({node.node_role}) has an unmapped role. Using its unflattened original configuration without creating a dataclass instance.")
            node.config = original_node_config_dict

    logger.info("TaskGraph node configuration update complete.")
    return workerflow_taskgraph


def display_node_config(workerflow_taskgraph: TaskGraph) -> None:
    """
    Prints the configuration for each node.
    This version is adapted for when node.config primarily holds a dataclass instance.
    """
    if not isinstance(workerflow_taskgraph, TaskGraph):
        logger.error("Error: Input must be a TaskGraph object.")
        return

    if not workerflow_taskgraph.nodes:
        logger.warning(f"Graph '{workerflow_taskgraph.graph_id}' has no nodes.")
        return

    logger.debug(f"Displaying configurations for all nodes in graph '{workerflow_taskgraph.graph_id}':")

    for node_id, node in workerflow_taskgraph.nodes.items():
        if not isinstance(node.config, dict):
            logger.warning(f"Node '{node_id}' config is not a dictionary. Skipping.")
            continue

        dataclass_obj = node.config.get(INTERN_CONFIG)

        if dataclass_obj and is_dataclass(dataclass_obj):
            config_for_display = asdict(dataclass_obj)

            # Include the node ID in the displayed configuration
            if NODE_ID in node.config:
                config_for_display[NODE_ID] = node.config[NODE_ID]

            log_dict_formatted(config_for_display, title=f"Node: {node_id} Configuration Details", log_level="debug")
        else:
            # If the config is not a dataclass, log the raw dictionary
            logger.warning(f"Node '{node_id}' does not contain a valid dataclass in '{INTERN_CONFIG}'.")
            log_dict_formatted(node.config, title=f"Node: {node_id} Raw Configuration Details", log_level="debug")



# Copyright (c) 2025, Shanghai Innovation Institute.  All rights reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     <url id="d0rb3ebacc47j8if2om0" type="url" status="parsed" title="Apache License, Version 2.0" wc="10467">http://www.apache.org/licenses/LICENSE-2.0</url>
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from queue import Queue
from typing import Any, Dict, Optional, Tuple
import yaml
import json
import os
import importlib
import ray
import sys
from loguru import logger


# --- 1. Abstract Base Class for Environment ---


async def get_ddp_world_size_rank(local_world_size, local_rank, local_parallel_size):
    ddp_world_size = local_world_size // local_parallel_size
    ddp_rank = local_rank // local_parallel_size
    return ddp_world_size, ddp_rank


class BaseEnvironment(ABC):
    """
    BaseEnvironment defines functions for users to implement
    """

    def __init__(self, env_id: str, config: Dict[str, Any]):
        self.env_id = env_id
        self.config = config
        self._validate_config()
        logger.info(f"[BaseEnvironment] Environment '{self.env_id}' initialized, configuration: {self.config}")

    def _validate_config(self):
        """(Optional) Validates whether the input configuration meets the environment requirements."""
        pass

    async def _reset(self, dp_rank: int, ddp_world_size: int, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None) -> Any:
        obs = await self.reset(dp_rank, ddp_world_size, seed, options)
        return await self.format_observation(obs)

    async def reset(self, dp_rank: int, ddp_world_size: int, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None) -> Any:
        """
        Resets the environment to its initial state.
        Args:
            seed: The seed used for environment randomness.
            options: Environment-specific reset options.
        Returns:
            Initial observation (raw observation).
        """
        logger.info("Reset not implemented")
        raise NotImplementedError
        return obs

    async def step(self, action: Any) -> Tuple[Any, Any, Dict[str, Any]]:
        """
        Executes an action in the environment. Subclasses can override this method. Alternatively, you can override do_action, get_rewards, and get_obs, but with limited flexibility.
        Args:
            action: Input parameters for executing the action.
        Returns:
            Tuple: (observation, reward, terminated, truncated, info)
                   - observation: The environment's raw observation.
                   - reward: The reward obtained after executing the action.
                   - info: A dictionary containing additional diagnostic information as needed.
        """
        logger.info("Step not implemented")
        raise NotImplementedError
        return next_obs, rewards, info

    async def get_rewards(self, actions: Any) -> Any:
        """Returns the rewards of action"""
        return None

    async def _step(self, action: Any) -> Tuple[Any, float, Dict[str, Any]]:
        """
        Adds env_id to the returned data after executing the action.
        """
        obs, rewards, info = await self.step(action)
        info["sii_env_id"] = self.env_id
        obs = await self.format_observation(obs)
        return obs, rewards, info

    async def format_observation(self, observation: Any) -> Any:
        """
        Formats the environment's raw observation into a format suitable for LLM input.
        The default implementation directly returns the raw observation. Subclasses can override this method.
        Automatically called by _step and _reset.
        Args:
            observation: The raw observation from step() or reset().
        Returns:
            Formatted observation.
        """
        # logger.info(f"[BaseEnvironment {self.env_id}] Formatting observation (default implementation).")
        return observation

    async def get_observation_space(self) -> Any:
        """Returns the definition of the environment's observation space (e.g., a gym.spaces.Space object)."""
        # raise NotImplementedError # Encouraged for subclasses to implement
        return None  # Simplified

    async def get_action_space(self) -> Any:
        """Returns the definition of the environment's action space."""
        # raise NotImplementedError # Encouraged for subclasses to implement
        return None  # Simplified

    async def close(self) -> None:
        """Cleans up environment resources if needed."""
        logger.info(f"[BaseEnvironment {self.env_id}] Environment closed.")
        pass


# --- 2. Ray Actor Wrapping Environment Instances ---
@ray.remote
class RayEnvRunner:
    def __init__(self, env_id: str, env_config: Dict[str, Any], env_class_ref: str):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        sys.path.append(current_dir)
        module_path, class_name = env_class_ref.rsplit(".", 1)
        module = importlib.import_module(module_path)
        env_class = getattr(module, class_name)
        if not issubclass(env_class, BaseEnvironment):
            raise TypeError(f"Class '{env_class_ref}' must be a subclass of BaseEnvironment.")
        self.env = env_class(env_id, env_config)

    def _validate_config(self):
        self.env._validate_config()

    async def reset(self, local_world_size: int, local_rank: int, local_parallel_size: int, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None) -> Any:
        ddp_world_size, dp_rank = await get_ddp_world_size_rank(local_world_size, local_rank, local_parallel_size)
        futures = await self.env._reset(dp_rank, ddp_world_size, seed, options)
        return futures

    async def step(self, action: Any) -> Tuple[Any, Any, Dict[str, Any]]:
        return await self.env._step(action)

    async def get_rewards(self, action: Any) -> Any:
        """Returns the rewards of action"""
        return await self.env.get_rewards(action)

    async def format_observation(self, observation: Any) -> Any:
        """
        Formats the environment's raw observation into a format suitable for LLM input.
        The default implementation directly returns the raw observation. Subclasses can override this method.
        Automatically called by _step and _reset.
        Args:
            observation: The raw observation from step() or reset().
        Returns:
            Formatted observation.
        """
        # logger.info(f"[BaseEnvironment {self.env_id}] Formatting observation (default implementation).")
        return await self.env.format_observation(observation)

    async def get_observation_space(self) -> Any:
        """Returns the definition of the environment's observation space (e.g., a gym.spaces.Space object)."""
        # raise NotImplementedError # Encouraged for subclasses to implement
        return await self.env.get_observation_space()  # Simplified

    async def get_action_space(self) -> Any:
        """Returns the definition of the environment's action space."""
        # raise NotImplementedError # Encouraged for subclasses to implement
        return await self.env.get_action_space()  # Simplified

    async def close(self) -> None:
        """Cleans up environment resources if needed."""
        return await self.env.close()  # Simplified


"""
this method requires env don't save status
"""


@ray.remote
class ParallelRayEnvRunner:
    """
    Multi Ray Actor used
    """

    def __init__(self, env_id: int, agent_parallel_num: int, env_class_ref: str, env_config: Dict[str, Any]):
        """
        Args:
            env_id: A unique identifier for the environment.
            env_class_ref: A reference string pointing to a subclass of BaseEnvironment (e.g., "my_module.MyCustomEnv").
            env_config: A dictionary of configurations passed to the environment constructor.
        """
        self.env_id = env_id
        self.env_worker = Queue()
        # todo: calculate how many workers are in one environment, now max = 4
        self.worker_num = agent_parallel_num
        try:
            for i in range(self.worker_num):
                self.env_worker.put(RayEnvRunner.remote(env_id=env_id, env_config=env_config, env_class_ref=env_class_ref))
            logger.info(f"[RayEnvActor {self.env_id}] Successfully instantiated environment type: {env_class_ref}")
        except Exception as e:
            logger.info(f"Error: [RayEnvActor {self.env_id}] Failed to initialize environment '{env_class_ref}': {e}")
            # In practical applications, a more robust error handling or retry mechanism may be needed
            raise

    async def reset(self, local_world_size: int, local_rank: int, local_parallel_size: int, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None) -> Any:
        # logger.info(f"[RayEnvActor {self.env_id}] Calling reset...")
        # Use all envWorkers to reset
        futures = []
        for i in range(self.worker_num):
            worker = self.env_worker.get()
            futures.append(worker.reset.remote(local_world_size, local_rank, local_parallel_size, seed, options))
            self.env_worker.put(worker)
        return futures

    async def step(self, action: Any) -> Any:
        # logger.info(f"[RayEnvActor {self.env_id}] Calling step, action: {action is not None}")
        worker = self.env_worker.get()
        futures = worker.step.remote(action)
        self.env_worker.put(worker)
        return futures

    async def get_rewards(self, action: Any) -> Any:
        # logger.info(f"[RayEnvActor {self.env_id}] Calling step, action: {action is not None}")
        worker = self.env_worker.get()
        futures = worker.get_rewards.remote(action)
        self.env_worker.put(worker)
        return futures

    async def format_observation(self, observation: Any) -> Any:
        # logger.info(f"[RayEnvActor {self.env_id}] Calling format_observation...")
        worker = self.env_worker.get()
        futures = worker.format_observation.remote(observation)
        self.env_worker.put(worker)
        return futures

    async def get_observation_space(self) -> Any:
        worker = self.env_worker.get()
        futures = worker.get_observation_space.remote()
        self.env_worker.put(worker)
        return futures

    async def get_action_space(self) -> Any:
        worker = self.env_worker.get()
        futures = worker.get_action_space.remote()
        self.env_worker.put(worker)
        return futures

    async def close(self) -> None:
        for i in range(self.worker_num):
            worker = self.env_worker.get()
            await worker.close.remote()

    def get_env_id(self) -> str:
        return self.env_id


# --- 3. Environment Configuration Loader ---
class EnvironmentConfigLoader:
    """
    Loads environment configurations from configuration files (YAML/JSON) and creates RayEnvironmentActor instances.
    """

    def __init__(self):
        pass

    @staticmethod
    def load_environments_from_file(file_path: str) -> Dict[int, "ray.actor.ActorHandle"]:
        """
        Loads all environment definitions from the specified configuration file and starts a RayEnvironmentActor for each definition.
        Args:
            file_path: The path to the YAML or JSON configuration file.
        Returns:
            A dictionary mapping env_id to the corresponding RayEnvironmentActor handle.
        """
        _, file_extension = os.path.splitext(file_path)
        file_extension = file_extension.lower()

        raw_configs = None
        try:
            with open(file_path, "r", encoding="utf-8") as f:
                if file_extension in [".yaml", ".yml"]:
                    raw_configs_list = yaml.safe_load(f)
                elif file_extension == ".json":
                    raw_configs_list = json.load(f)
                else:
                    raise ValueError(f"Unsupported configuration file extension: {file_extension}")

            if not isinstance(raw_configs_list, list):  # Assume the configuration file is a list of environment configuration objects
                raise ValueError("The environment configuration file should contain a list of environment configuration objects.")

        except Exception as e:
            logger.info(f"Error: Failed to load or parse environment configuration file '{file_path}': {e}")
            raise

        environment_actors: Dict[int, "ray.actor.ActorHandle"] = {}
        for env_conf in raw_configs_list:
            env_id = env_conf.get("environment_id", None)
            env_class_ref = env_conf.get("environment_type_ref", None)
            env_specific_config = env_conf.get("config", {})
            ray_actor_options = env_conf.get("source_options", {})  # E.g., {"num_cpus": 1}

            if env_id is None or env_class_ref is None:
                logger.info(f"Warning: Skipping invalid environment configuration (missing id or type_ref): {env_conf}")
                continue
            try:
                agent_parallel_num = env_conf.get("agent_parallel_num", 1)
                n_agent = env_specific_config.get("num_agents", 1)
                if agent_parallel_num > 1 and n_agent > 1:
                    actor_handle = ParallelRayEnvRunner.options(**ray_actor_options).remote(env_id=env_id, agent_parallel_num=min(agent_parallel_num, n_agent), env_class_ref=env_class_ref, env_config=env_specific_config)
                    environment_actors[env_id] = actor_handle
                    logger.info(f"Successfully started RayEnvironmentActor for env_id='{env_id}' (type: {env_class_ref}).")
                else:
                    actor_handle = RayEnvRunner.options(**ray_actor_options).remote(env_id=env_id, env_class_ref=env_class_ref, env_config=env_specific_config)
                    environment_actors[env_id] = actor_handle
                    logger.info(f"Successfully started RayEnvironmentActor for env_id='{env_id}' (type: {env_class_ref}).")
            except Exception as e:
                logger.info(f"Error: Failed to start RayEnvironmentActor for env_id='{env_id}': {e}")

        return environment_actors



# Copyright 2025, Shanghai Innovation Institute. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from collections import defaultdict
from typing import Any, Dict, List, Optional

import ray
from loguru import logger
from torch.distributed import ProcessGroup

from siirl.models.loader import TokenizerModule
from siirl.workers.base_worker import Worker
from siirl.scheduler.process_group_manager import ProcessGroupManager
from siirl.utils.params import SiiRLArguments
from siirl.workers.dag import TaskGraph
from siirl.workers.databuffer import DataProto

from .constants import DAGInitializationError
from .mixins.execution_mixin import ExecutionMixin
from .mixins.initialization_mixin import InitializationMixin
from .mixins.node_executors_mixin import NodeExecutorsMixin
from .mixins.utilities_mixin import UtilitiesMixin
from .mixins.validation_mixin import ValidationMixin


class DAGWorker(InitializationMixin, ExecutionMixin, NodeExecutorsMixin, ValidationMixin, UtilitiesMixin, Worker):
    """
    Orchestrates a Directed Acyclic Graph (DAG) of tasks for distributed training,
    managing the setup, initialization, and workflow for a specific rank.
    """

    def __init__(
        self,
        config: SiiRLArguments,
        process_group_manager: ProcessGroupManager,
        taskgraph_mapping: Dict[int, TaskGraph],
        data_buffers: List["ray.actor.ActorHandle"],
        environments: Optional[Dict[int, "ray.actor.ActorHandle"]] = None,
    ):
        super().__init__()
        self.config = config
        self.process_group_manager = process_group_manager
        self.taskgraph_mapping = taskgraph_mapping
        self.data_buffers = data_buffers
        self.environments = environments
        self.enable_perf = os.environ.get("SIIRL_ENABLE_PERF", "0") == "1" or config.dag.enable_perf

        # State attributes
        self.global_steps = 0
        self.total_training_steps = 0
        self.workers: Dict[str, Any] = {}
        self.agent_group_worker: Dict[int, Dict["NodeRole", Any]] = defaultdict(dict)
        self.agent_group_process_group: Dict[int, Dict["NodeRole", Any]] = defaultdict(dict)
        self.process_groups: Dict[str, ProcessGroup] = {}
        self.tokenizer_mapping: Dict[str, TokenizerModule] = {}
        self.kl_ctrl_in_reward = None
        self.logger = None
        self.progress_bar = None
        self._rank: int = -1
        self.taskgraph: Optional[TaskGraph] = None
        self.internal_data_cache: Dict[str, DataProto] = {}
        self.agent_critic_worker: Any
        # Finish flag
        self.taskgraph_execute_finished = False

        # async rollout
        self.rollout_mode = 'sync'
        self._async_rollout_manager = None
        self.zmq_address = None # used for async_vllmrollout

        try:
            self._initialize_worker()
        except (ValueError, TypeError, KeyError, AttributeError, NotImplementedError) as e:
            rank = os.environ.get("RANK", "UNKNOWN")
            logger.error(f"Rank {rank}: Failed to create DAGWorker due to a critical setup error: {e}", exc_info=True)
            raise DAGInitializationError(f"Initialization failed on Rank {rank}: {e}") from e

        self.log_ray_actor_info()



from siirl.workers.databuffer import DataProto
from siirl.workers.dag.node import Node
from tensordict import TensorDict


def add_prefix_to_dataproto(data_proto: DataProto, node: Node):
    """
    Adds a prefix to all keys in the DataProto.
    The prefix is formatted as f"agent_group_{node.agent_group}_".
    Only keys that do not already have a prefix will be modified.

    Args:
        data_proto (DataProto): The DataProto instance.
        node (Node): The node containing the agent_group.
    """
    prefix = f"agent_group_{node.agent_group}_"
    prefix_agent_group = "agent_group_"

    # Process tensor batch
    if data_proto.batch is not None:
        new_batch = {}
        for key, value in data_proto.batch.items():
            if not key.startswith(prefix_agent_group):
                new_key = prefix + key
                new_batch[new_key] = value
            else:
                new_batch[key] = value
        data_proto.batch = TensorDict(new_batch, batch_size=data_proto.batch.batch_size)

    # Process non_tensor_batch
    if data_proto.non_tensor_batch is not None:
        new_non_tensor = {}
        for key, value in data_proto.non_tensor_batch.items():
            if not key.startswith(prefix_agent_group):
                new_key = prefix + key
                new_non_tensor[new_key] = value
            else:
                new_non_tensor[key] = value
        data_proto.non_tensor_batch = new_non_tensor

    # Process meta_info
    if data_proto.meta_info is not None:
        new_meta = {}
        for key, value in data_proto.meta_info.items():
            if not key.startswith(prefix_agent_group):
                new_key = prefix + key
                new_meta[new_key] = value
            else:
                new_meta[key] = value
        data_proto.meta_info = new_meta
    return data_proto


def remove_prefix_from_dataproto(data_proto, node: Node):
    """
    Removes the prefix from all keys in the DataProto.
    Only keys with a matching prefix will have the prefix removed.

    Args:
        data_proto (DataProto): The DataProto instance.
        node (Node): The node containing the agent_group to identify the prefix.
    """
    prefix = f"agent_group_{node.agent_group}_"
    prefix_len = len(prefix)

    # Process tensor batch
    if data_proto.batch is not None:
        new_batch = {}
        for key, value in data_proto.batch.items():
            if key.startswith(prefix):
                new_key = key[prefix_len:]
                new_batch[new_key] = value
            else:
                new_batch[key] = value
        data_proto.batch = TensorDict(new_batch, batch_size=data_proto.batch.batch_size)

    # Process non_tensor_batch
    if data_proto.non_tensor_batch is not None:
        new_non_tensor = {}
        for key, value in data_proto.non_tensor_batch.items():
            if key.startswith(prefix):
                new_key = key[prefix_len:]
                new_non_tensor[new_key] = value
            else:
                new_non_tensor[key] = value
        data_proto.non_tensor_batch = new_non_tensor

    # Process meta_info
    if data_proto.meta_info is not None:
        new_meta = {}
        for key, value in data_proto.meta_info.items():
            if key.startswith(prefix):
                new_key = key[prefix_len:]
                new_meta[new_key] = value
            else:
                new_meta[key] = value
        data_proto.meta_info = new_meta

    return data_proto
